Package: CBTF
Type: Package
Title: Caught by the Fuzz! - A Minimalistic Fuzz-Test Runner
Version: 0.7.0
Date: 2026-07-03
Authors@R: c(person("Marco", "Colombo",
                    email = "mar.colombo13@gmail.com", role = c("aut", "cre"),
                    comment = c(ORCID = "0000-0001-6672-0623")))
Maintainer: Marco Colombo <mar.colombo13@gmail.com>
Description: A simple runner for fuzz-testing functions in an R package's
    public interface. Fuzz testing helps identify functions lacking sufficient
    argument validation, and uncovers problematic inputs that, while valid by
    function signature, may cause issues within the function body.
URL: https://mcol.github.io/caught-by-the-fuzz/
BugReports: https://github.com/mcol/caught-by-the-fuzz/issues
Imports: mirai (>= 2.5.2), rlang (>= 1.1.7), cli (>= 3.6.5)
Suggests: testthat (>= 3.2.3), withr (>= 3.0.2)
License: GPL-3
Encoding: UTF-8
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-03 13:20:49 UTC; marco
Author: Marco Colombo [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6672-0623>)
Repository: CRAN
Date/Publication: 2026-07-03 13:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("[[",cbtf)
S3method(length,cbtf)
S3method(print,cbtf)
S3method(summary,cbtf)
export(fuzz)
export(get_exported_functions)
export(test_inputs)
export(whitelist)

# CBTF: Caught by the Fuzz! A minimalistic fuzz-test runner for R<a href="https://www.youtube.com/watch?v=uJ-mpul94eo"><img src="man/figures/logo.png" align="right" height="120" /></a>

<!-- badges: start -->

[![CRAN](https://www.r-pkg.org/badges/version/CBTF)](https://CRAN.R-project.org/package=CBTF)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
<!-- badges: end -->

The `CBTF` package implements a very simple mechanism for fuzz-testing
functions in the public interface of an R package.

Fuzz testing helps identify functions that lack sufficient argument
validation, and uncovers sets of inputs that, while valid by function
signature, may cause issues within the function body.

The core functionality of the package is in the `fuzz()` function, which
calls each provided function with a certain input and records the output
produced. If an error or a warning is generated, this is captured and
reported to the user, unless it matches a pattern of whitelisted
messages, as specified in the `ignore_patterns` argument. The objects
returned by `fuzz()` can be inspected with `summary()` and `print()`.

Whitelisting can also be done after a fuzz run has been completed via
the `whitelist()` function, so that only messages that need to be acted
upon are actually shown. Using `whitelist()` has the advantage of not
requiring the completion of a fuzz run of all functions over all inputs
again.

Note that `fuzz()` relies on the [`mirai`
package](https://mirai.r-lib.org/) for asynchronous operations and
parallelisation, and execution occurs on persistent background
processes. These can be started automatically by specifying the
`daemons` option; alternatively, they can be set up manually with the
`mirai::daemons()` function; refer to the original `mirai` documentation
for a complete description of its arguments and behaviour.

The helper function `get_exported_functions()` identifies the functions
in the public interface of a given package, facilitating the generation
of the list of functions to be fuzzed.

The helper function `test_inputs()` is invoked by `fuzz()` if the user
doesn’t specify the set of inputs to be tested. By default it generates
a large set of potentially problematic inputs, but these can be limited
just to the desired classes of inputs.

A write-up on the background of the development of `CBTF` is available
in [this post on the REPLAY
website](https://replay.geog.uni-heidelberg.de/REPLAY-website/post/2025/08/fuzzing-luminescence-to-improve-it/).

## Usage

This is a simple example that displays how to use `CBTF` to fuzz an R
package. We consider `mime` because it is small enough to run quickly
and is likely installed on most systems.

``` r
library(CBTF)
funs <- get_exported_functions("mime")
(res <- fuzz(funs, what = list(TRUE)))
```

    ## ℹ Fuzzing 2 functions with 1 input (using 2 daemons)
    ## ℹ 2 tests run  [4ms]
    ## ✖  🚨   CAUGHT BY THE FUZZ!   🚨
    ## 
    ## ── Test input [[1]]: TRUE
    ##       guess_type  FAIL  a character vector argument expected
    ##  parse_multipart  FAIL  $ operator is invalid for atomic vectors
    ## 
    ##  [ FAIL 2 | WARN 0 | SKIP 0 | OK 0 ]

The first occurrence is a false positive, as the message returned
indicates that the input was checked and the function returned cleanly.
The second case instead reveals that the function didn’t validate its
input: indeed, it expected an environment, and used the `$` operation on
it without checking.

The false positive result can be easily removed by whitelisting an
appropriate pattern:

``` r
whitelist(res, "a character vector argument expected")
```

    ## ✖  🚨   CAUGHT BY THE FUZZ!   🚨
    ## 
    ## ── Test input [[1]]: TRUE
    ##  parse_multipart  FAIL  $ operator is invalid for atomic vectors
    ## 
    ##  [ FAIL 1 | WARN 0 | SKIP 0 | OK 1 ]

For real-life application of `CBTF`, see [Luminescence issue
439](https://github.com/R-Lum/Luminescence/issues/439) and the bugs
linked from there.

### Fuzzing multiple arguments

The example above didn’t specify the `args` argument, and therefore only
the first function argument was fuzzed. By setting `args`, it’s possible
to provide a list of arguments (of any length) to pass to the functions
being fuzzed.

To fuzz the first three arguments of `matrix()` (`data`, `nrow` and
`ncol`), we could do the following:

``` r
fuzz("matrix", what = list(NA, NULL), args = list(1:4, 2, 2))
```

    ## ℹ Fuzzing 1 function with 6 inputs (using 2 daemons)
    ## ℹ Functions will be searched in the global namespace as `package` was not specified
    ## ℹ 6 tests run  [6ms]
    ## ✖  🚨   CAUGHT BY THE FUZZ!   🚨
    ## 
    ## ── Test input [[2]]: NULL, 2, 2
    ##  matrix  FAIL  'data' must be of a vector type, was 'NULL'
    ## 
    ## ── Test input [[3]]: 1:4, NA, 2
    ##  matrix  FAIL  invalid 'nrow' value (too large or NA)
    ## 
    ## ── Test input [[5]]: 1:4, 2, NA
    ##  matrix  FAIL  invalid 'ncol' value (too large or NA)
    ## 
    ##  [ FAIL 3 | WARN 0 | SKIP 0 | OK 3 ]

If names are given to elements in `args`, they will be used in the input
lists generated. This can be helpful to fuzz a specific argument across
several function, or to fuzz an argument that comes later in the
argument list or that are passed via `...`.

For example, the following will fuzz the `data`, `nrow` and `dimnames`
arguments:

``` r
fuzz("matrix", what = list(NA, NULL), args = list(1:4, 2, dimnames = NULL))
```

    ## ℹ Fuzzing 1 function with 6 inputs (using 2 daemons)
    ## ℹ Functions will be searched in the global namespace as `package` was not specified
    ## ℹ 6 tests run  [6ms]
    ## ✖  🚨   CAUGHT BY THE FUZZ!   🚨
    ## 
    ## ── Test input [[2]]: NULL, 2, dimnames = NULL
    ##  matrix  FAIL  'data' must be of a vector type, was 'NULL'
    ## 
    ## ── Test input [[3]]: 1:4, NA, dimnames = NULL
    ##  matrix  FAIL  invalid 'nrow' value (too large or NA)
    ## 
    ## ── Test input [[5]]: 1:4, 2, dimnames = NA
    ##  matrix  FAIL  'dimnames' must be a list
    ## 
    ##  [ FAIL 3 | WARN 0 | SKIP 0 | OK 3 ]

Fuzzing is generally more effective when the values in `args` provide a
good default for the functions being fuzzed; while an argument is being
fuzzed, the others remain valid, allowing for a broader exploration of
the argument space.

However, this is not strictly necessary (and may not be achievable when
fuzzing multiple functions simultaneously). Additionally, some code
paths may only be reached through combinations of invalid values that
would otherwise remain unexplored.

## Advanced topics

### Generated argument lists

Either `args` or `what` can be `NULL`, but not both. The most basic use
sets `args = NULL`, in which case each of the inputs in `what` is
assigned to the first argument of each function being fuzzed. For
example, if `what = list(NA, "")`, then the following input lists will
be generated:

- `list(NA)`
- `list("")`

If instead `what = NULL`, the exact argument list defined in `args` will
be tested on all functions without any fuzzing. This can be useful after
a fuzz run to collect results on a specific set of problematic inputs,
or to monitor progress while addressing the bugs discovered during the
fuzzing process.

However, the full potential of `CBTF` is realised by specifying both
`what` (or leaving it unset, which deploys the full range of inputs
generated by `test_inputs()`) and `args`. In such cases, `args` is
expanded internally into multiple lists, each differing by one elements
modified according to an element in `what`.

For example, if `what = list(NA, "")` and
`args = list(x = 11, y = 22, 33)`, the following input lists will be
generated and tested:

- `list(x = NA, y = 22, 33)`
- `list(x = "", y = 22, 33)`
- `list(x = 11, y = NA, 33)`
- `list(x = 11, y = "", 33)`
- `list(x = 11, y = 22, NA)`
- `list(x = 11, y = 22, "")`

Sometimes it may be helpful to specify an argument to have a fixed
value, so that it remains unchanged while fuzzing. This can be done by
prepending “..” to the argument name. For example, to fix argument `x`
at 11, we would use `args = list(..x = 11, y = 22, 33)`, and only these
inputs will be tested:

- `list(x = 11, y = NA, 33)`
- `list(x = 11, y = "", 33)`
- `list(x = 11, y = 22, NA)`
- `list(x = 11, y = 22, "")`

### Parallel execution

The implementation of `fuzz()` uses the [`mirai`
package](https://mirai.r-lib.org/) for asynchronous operations and
parallelisation. This allows to fuzz the functions on persistent
background processes (daemons) in parallel.

There are two main approaches to control parallel execution:

1.  Setting the `daemons` argument to an integer value (2 by default):
    this will start as many daemons as specified (and shut them down
    automatically at the end of the fuzz run). Note that there is no
    benefit in starting more daemons than the number of available cores.

    ``` r
    ## this will start 4 daemons
    res <- fuzz(funs, daemons = 4)
    ```

2.  Manually setting up the daemons before the start of the function:
    this can be accomplished via `mirai::daemons()`, which allows to
    specify remote daemons as well as local one. This also avoids the
    cost of starting and closing daemons if `fuzz()` were to be called
    multiple times. It remains responsibility of the user to close the
    daemons when no longer in use. When active daemons are found, the
    `daemons` argument will be ignored. Refer to the original `mirai`
    documentation for a complete description of its arguments and
    behaviour.

    ``` r
    ## set up persistent background processes on the local machine
    mirai::daemons(4)
    res <- fuzz(funs)

    ## close the background processes
    mirai::daemons(0)
    ```

### Timeouts

Long-running functions can slow down the progress of `fuzz()`, so by
default if a function does not produce an error within 2 seconds, it
will be stopped. A timed out function returns an `"OK"` result, with the
corresponding `$msg` field recording that a timeout was applied.

However, the default timeout may be too short (or perhaps too long) in
some applications. If desired, the maximum running time of a job (in
seconds) can be controlled via the `timeout` argument of `fuzz()`.

### Controlling the inputs tested

By default, `fuzz()` tests all the inputs produced by `test_inputs()`
(currently 85 inputs). However, this can be controlled by specifying the
classes that should be tested:

``` r
test_inputs(use = c("scalar", "numeric", "integer", "matrix"))
```

Alternatively, one can specify the classes to be excluded:

``` r
test_inputs(skip = c("date", "raw"))
```

A vector of valid classes can be retrieved programmatically by setting
this argument to “help”:

``` r
test_inputs("help")
```

    ##  [1] "all"        "scalar"     "numeric"    "integer"    "logical"   
    ##  [6] "character"  "factor"     "data.frame" "matrix"     "array"     
    ## [11] "date"       "time"       "raw"        "na"         "list"

### Fuzzing list inputs automatically

It is trivial to augment a given set of inputs with list versions of the
same. This effectively doubles the number of tests run with no
additional coding effort.

``` r
fuzz(funs, what = list(letters), listify_what = TRUE)
```

    ## ℹ Fuzzing 2 functions with 2 inputs (using 2 daemons)
    ## ℹ 4 tests run  [5ms]
    ## ✖  🚨   CAUGHT BY THE FUZZ!   🚨
    ## 
    ## ── Test input [[1]]: letters
    ##  parse_multipart  FAIL  $ operator is invalid for atomic vectors
    ## 
    ## ── Test input [[2]]: list(letters)
    ##       guess_type  FAIL  a character vector argument expected
    ## 
    ##  [ FAIL 2 | WARN 0 | SKIP 0 | OK 2 ]

## Funding

Development of `CBTF` is partially supported through the DFG programme
“REPLAY: REProducible Luminescence Data AnalYses” [No
528704761](https://gepris.dfg.de/project/528704761) led by Dr Sebastian
Kreutzer (PI at LIAG - Institute for Applied Geophysics, Hannover, DE)
and Dr Thomas Kolb (PI at Justus-Liebig-University Giessen, DE). Updates
on the REPLAY project at large are available at the [REPLAY
website](https://replay.geog.uni-heidelberg.de/).

# CBTF: Caught by the Fuzz! A minimalistic fuzz-test runner for R<a href="https://www.youtube.com/watch?v=uJ-mpul94eo"><img src="man/figures/logo.png" align="right" height="120" /></a>

<!-- badges: start -->

[![CRAN](https://www.r-pkg.org/badges/version/CBTF)](https://CRAN.R-project.org/package=CBTF)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
<!-- badges: end -->

The `CBTF` package implements a very simple mechanism for fuzz-testing
functions in the public interface of an R package.

Fuzz testing helps identify functions that lack sufficient argument
validation, and uncovers sets of inputs that, while valid by function
signature, may cause issues within the function body.

The core functionality of the package is in the `fuzz()` function, which
calls each provided function with a certain input and records the output
produced. If an error or a warning is generated, this is captured and
reported to the user, unless it matches a pattern of whitelisted
messages, as specified in the `ignore_patterns` argument. The objects
returned by `fuzz()` can be inspected with `summary()` and `print()`.

Whitelisting can also be done after a fuzz run has been completed via
the `whitelist()` function, so that only messages that need to be acted
upon are actually shown. Using `whitelist()` has the advantage of not
requiring to run the fuzzer over all functions and all inputs again.

Note that `fuzz()` uses the [`mirai` package](https://mirai.r-lib.org/)
for asynchronous operations and parallelisation, and execution occurs on
persistent background processes. These can be started automatically by
specifying the `daemons` option; alternatively, they can be set up
manually with the `mirai::daemons()` function; refer to the original
`mirai` documentation for a complete description of its arguments and
behaviour.

The helper function `get_exported_functions()` identifies the functions
in the public interface of a given package, facilitating the generation
of the list of functions to be fuzzed.

The helper function `test_inputs()` is invoked by `fuzz()` if the user
doesn’t specify the set of inputs to be tested. By default it generates
a large set of potentially problematic inputs, but these can be limited
just to the desired classes of inputs.

## Usage

This is a simple example that displays how to use `CBTF` to fuzz an R
package. We consider `mime` because it is small enough to run quickly
and is likely installed on most systems.

``` r
library(CBTF)
funs <- get_exported_functions("mime")
(res <- fuzz(funs, what = list(TRUE)))
```

    ## ℹ Fuzzing 2 functions with 1 input (using 2 daemons)
    ## ℹ 2 tests run  [4ms]
    ## ✖  🚨   CAUGHT BY THE FUZZ!   🚨
    ## 
    ## ── Test input [[1]]: TRUE
    ##       guess_type  FAIL  a character vector argument expected
    ##  parse_multipart  FAIL  $ operator is invalid for atomic vectors
    ## 
    ##  [ FAIL 2 | WARN 0 | SKIP 0 | OK 0 ]

The first occurrence is a false positive, as the message returned
indicates that the input was checked and the function returned cleanly.
The second case instead reveals that the function didn’t validate its
input: indeed, it expected an environment, and used the `$` operation on
it without checking.

The false positive result can be easily removed by whitelisting an
appropriate pattern:

``` r
whitelist(res, "a character vector argument expected")
```

    ## ✖  🚨   CAUGHT BY THE FUZZ!   🚨
    ## 
    ## ── Test input [[1]]: TRUE
    ##  parse_multipart  FAIL  $ operator is invalid for atomic vectors
    ## 
    ##  [ FAIL 1 | WARN 0 | SKIP 0 | OK 1 ]

### Fuzzing multiple arguments

The example above didn’t specify the `args` argument, and therefore only
the first function argument was fuzzed. By setting `args`, it’s possible
to provide a list of arguments (of any length) to pass to the functions
being fuzzed.

To fuzz the first three arguments of `matrix()` (`data`, `nrow` and
`ncol`), we could do the following:

``` r
fuzz("matrix", what = list(NA, NULL), args = list(1:4, 2, 2))
```

    ## ℹ Fuzzing 1 function with 6 inputs (using 2 daemons)
    ## ℹ Functions will be searched in the global namespace as `package` was not specified
    ## ℹ 6 tests run  [6ms]
    ## ✖  🚨   CAUGHT BY THE FUZZ!   🚨
    ## 
    ## ── Test input [[2]]: NULL, 2, 2
    ##  matrix  FAIL  'data' must be of a vector type, was 'NULL'
    ## 
    ## ── Test input [[3]]: 1:4, NA, 2
    ##  matrix  FAIL  invalid 'nrow' value (too large or NA)
    ## 
    ## ── Test input [[5]]: 1:4, 2, NA
    ##  matrix  FAIL  invalid 'ncol' value (too large or NA)
    ## 
    ##  [ FAIL 3 | WARN 0 | SKIP 0 | OK 3 ]

If names are given to elements in `args`, they will be used in the input
lists generated. This can be helpful to fuzz a specific argument across
several function, or to fuzz an argument that comes later in the
argument list or that are passed via `...`.

For example, the following will fuzz the `data`, `nrow` and `dimnames`
arguments:

``` r
fuzz("matrix", what = list(NA, NULL), args = list(1:4, 2, dimnames = NULL))
```

    ## ℹ Fuzzing 1 function with 6 inputs (using 2 daemons)
    ## ℹ Functions will be searched in the global namespace as `package` was not specified
    ## ℹ 6 tests run  [6ms]
    ## ✖  🚨   CAUGHT BY THE FUZZ!   🚨
    ## 
    ## ── Test input [[2]]: NULL, 2, dimnames = NULL
    ##  matrix  FAIL  'data' must be of a vector type, was 'NULL'
    ## 
    ## ── Test input [[3]]: 1:4, NA, dimnames = NULL
    ##  matrix  FAIL  invalid 'nrow' value (too large or NA)
    ## 
    ## ── Test input [[5]]: 1:4, 2, dimnames = NA
    ##  matrix  FAIL  'dimnames' must be a list
    ## 
    ##  [ FAIL 3 | WARN 0 | SKIP 0 | OK 3 ]

Fuzzing is generally more effective when the values in `args` provide a
good default for the functions being fuzzed; while an argument is being
fuzzed, the others remain valid, allowing for a broader exploration of
the argument space.

However, this is not strictly necessary (and may not be achievable when
fuzzing multiple functions simultaneously). Additionally, some code
paths may only be reached through combinations of invalid values that
would otherwise remain unexplored.

## Advanced topics

### Generated argument lists

Either `args` or `what` can be `NULL`, but not both. The most basic use
sets `args = NULL`, in which case each of the inputs in `what` is
assigned to the first argument of each function being fuzzed. For
example, if `what = list(NA, "")`, then the following input lists will
be generated:

- `list(NA)`
- `list("")`

If instead `what = NULL`, the exact argument list defined in `args` will
be tested on all functions without any fuzzing. This can be useful after
a fuzz run to collect results on a specific set of problematic inputs,
or to monitor progress while addressing the bugs discovered during the
fuzzing process.

However, the full potential of `CBTF` is realised by specifying both
`what` (or leaving it unset, which deploys the full range of inputs
generated by `test_inputs()`) and `args`. In such cases, `args` is
expanded internally into multiple lists, each differing by one elements
modified according to an element in `what`.

For example, if `what = list(NA, "")` and
`args = list(x = 11, y = 22, 33)`, the following input lists will be
generated and tested:

- `list(x = NA, y = 22, 33)`
- `list(x = "", y = 22, 33)`
- `list(x = 11, y = NA, 33)`
- `list(x = 11, y = "", 33)`
- `list(x = 11, y = 22, NA)`
- `list(x = 11, y = 22, "")`

Sometimes it may be helpful to specify an argument ho have a fixed
value, so that it remains unchanged while fuzzing. This can be done by
prepending “..” to the argument name. For example, to fix argument `x`
at 11, we would use `args = list(..x = 11, y = 22, 33)`, and only these
inputs will be tested:

- `list(x = 11, y = NA, 33)`
- `list(x = 11, y = "", 33)`
- `list(x = 11, y = 22, NA)`
- `list(x = 11, y = 22, "")`

### Parallel execution

The implementation of `fuzz()` uses the [`mirai`
package](https://mirai.r-lib.org/) for asynchronous operations and
parallelisation. This allows to fuzz the functions on persistent
background processes (daemons) in parallel.

There are two main approaches to control parallel execution:

1.  Setting the `daemons` argument to an integer value (2 by default):
    this will start as many daemons as specified (and shut them down
    automatically at the end of the fuzz run). Note that there is no
    benefit in starting more daemons than the number of available cores.

    ``` r
    ## this will start 4 daemons
    res <- fuzz(funs, daemons = 4)
    ```

2.  Manually setting up the daemons before the start of the function:
    this can be accomplished via `mirai::daemons()`, which allows to
    specify remote daemons as well as local one. This also avoids the
    cost of starting and closing daemons if `fuzz()` were to be called
    multiple times. It remains responsibility of the user to close the
    daemons when no longer in use. When active daemons are found, the
    `daemons` argument will be ignored. Refer to the original `mirai`
    documentation for a complete description of its arguments and
    behaviour.

    ``` r
    ## set up persistent background processes on the local machine
    mirai::daemons(4)
    res <- fuzz(funs)

    ## close the background processes
    mirai::daemons(0)
    ```

### Timeouts

Long-running functions can slow down the progress of `fuzz()`, so by
default if a function does not produce an error within 2 seconds, it
will be stopped. A timed out function returns an `"OK"` result, with the
corresponding `$msg` field recording that a timeout was applied.

However, the default timeout may be too short (or perhaps too long) in
some applications. If desired, the maximum running time of a job (in
seconds) can be controlled via the `timeout` argument of `fuzz()`

### Controlling the inputs tested

By default, `fuzz()` tests all the inputs produced by `test_inputs()`
(currently 70 inputs). However, this can be controlled by specifying the
classes that should be tested:

``` r
test_inputs(use = c("scalar", "numeric", "integer", "matrix"))
```

Alternatively, one can specify the classes to be excluded:

``` r
test_inputs(skip = c("date", "raw"))
```

A vector of valid classes can be retrieved programmatically by setting
this argument to “help”:

``` r
test_inputs("help")
```

    ##  [1] "all"        "scalar"     "numeric"    "integer"    "logical"   
    ##  [6] "character"  "factor"     "data.frame" "matrix"     "array"     
    ## [11] "date"       "raw"        "na"         "list"

### Fuzzing list inputs automatically

It is trivial to augment a given set of inputs with list versions of the
same. This effectively doubles the number of tests run with no
additional coding effort.

``` r
fuzz(funs, what = list(letters), listify_what = TRUE)
```

    ## ℹ Fuzzing 2 functions with 2 inputs (using 2 daemons)
    ## ℹ 4 tests run  [5ms]
    ## ✖  🚨   CAUGHT BY THE FUZZ!   🚨
    ## 
    ## ── Test input [[1]]: letters
    ##  parse_multipart  FAIL  $ operator is invalid for atomic vectors
    ## 
    ## ── Test input [[2]]: list(letters)
    ##       guess_type  FAIL  a character vector argument expected
    ## 
    ##  [ FAIL 2 | WARN 0 | SKIP 0 | OK 2 ]

## Funding

Development of `CBTF` is partially supported through the DFG programme
“REPLAY: REProducible Luminescence Data AnalYses” [No
528704761](https://gepris.dfg.de/gepris/projekt/528704761?language=en)
led by Dr Sebastian Kreutzer (PI at LIAG - Institute for Applied
Geophysics, Hannover, DE) and Dr Thomas Kolb (PI at
Justus-Liebig-University Giessen, DE). Updates on the REPLAY project at
large are available at the [REPLAY
website](https://replay.geog.uni-heidelberg.de/).

##===========================================================================
##
## Copyright (c) 2025-2026 Marco Colombo
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.
##
##===========================================================================

#' Results summary from a fuzz run
#'
#' Reports some summary statistics from the results of a run of [fuzz].
#'
#' The use of unicode icons in the output messages can be disabled by setting
#' `options(cli.unicode = FALSE)`.
#'
#' @param object An object of class `cbtf`.
#' @param tabulate Whether a tabulation of results should be printed out
#'        (`TRUE` by default). The tabulation can always be retrieved from
#'        the `"summary_table"` attribute of the returned object also when
#'        `tabulate = FALSE`.
#' @param ... Further arguments passed to or from other methods.
#'        These are currently ignored.
#'
#' @return
#' A data frame containing the following columns and attributes is returned
#' invisibly:
#' \item{fun}{The names of the function tested.}
#' \item{what}{The inputs tested.}
#' \item{res}{One of "OK", "FAIL", "WARN" or "SKIP" for each combination of
#'       function and input tested (see the *Value* section in [fuzz]).}
#' \item{msg}{The message received in case of error, warning or skip,
#'       or an empty string if no failure occurred.}
#' \item{attr(*, "summary_table")}{The tabulation of results that was printed
#'       out.}
#'
#' @examples
#' res <- fuzz(funs = c("list", "matrix", "mean"),
#'             what = test_inputs(c("numeric", "raw")))
#' summary(res)
#'
#' @seealso [print.cbtf]
#'
#' @export
summary.cbtf <- function(object, tabulate = TRUE, ...) {
  validate_class(tabulate, "logical", scalar = TRUE, from = "summary")
  df <- cbind(do.call(rbind, object$runs),
              fun = rep(object$funs, length(object$runs)),
              what = sapply(object$runs, function(x) attr(x, "what")))
  cli::cli_text("Fuzzed {nrow(object$runs[[1]])} function{?s} ",
                "on {length(object$runs)} input{?s}: ")
  tbl <- table(df$fun,
               factor(df$res, levels = c("FAIL", "WARN", "SKIP", "OK")))
  if (tabulate)
    print(tbl)
  cat("\n", compute_summary_stats(object, verbose = FALSE), "\n", sep = "")
  invisible(structure(df[, c("fun", "what", "res", "msg")],
                      summary_table = tbl))
}

#' Print the results from a fuzz run
#'
#' This formats the results from a fuzz run with colours and prints them to
#' the terminal. By default, only tests labelled as `"FAIL"` or `"WARN"` are
#' displayed, but this can be controlled via the `show` argument.
#'
#' The use of unicode icons in the output messages can be disabled by setting
#' `options(cli.unicode = FALSE)`.
#'
#' @param x An object of class `cbtf`.
#' @param show A character vector representing the subset of results to be
#'        printed, any of "fail", "warn", "skip", "ok", "all" and "none".
#' @param group Either `"input"` to show results grouped by test input
#'        (default), or `"function"` to show them grouped by function name.
#' @param ... Further arguments passed to or from other methods.
#'        These are currently ignored.
#'
#' @return
#' No return value, called for side effects.
#'
#' @examples
#' res <- fuzz(funs = c("list", "matrix", "mean"),
#'             what = test_inputs(c("numeric", "raw")))
#' print(res)
#' print(res, show = "all")
#' print(res, show = "none")
#' print(res, group = "function")
#'
#' @seealso [summary.cbtf]
#'
#' @export
print.cbtf <- function(x, show = c("fail", "warn"), group = "input", ...) {
  validate_class(show, "character", from = "print")
  validate_class(group, "character", scalar = TRUE,
                 choices = c("input", "function"), from = "print")
  show <- tolower(show)
  if ("none" %in% show) {
    cat(compute_summary_stats(x, verbose = FALSE), "\n")
    return(invisible())
  }
  if ("all" %in% show)
    show <- c("fail", "warn", "skip", "ok")
  summary.stats <- compute_summary_stats(x)
  res.size <- nchar(tocolour("FAIL")) # include ANSI formatting, if present

  if (group == "function") {
    ## Build a long-format data frame with all results
    df <- do.call(rbind, lapply(x$runs, function(run) {
      cbind(run, fun = x$funs, what = attr(run, "what"))
    }))

    for (fun in x$funs) {
      sub <- df[df$fun == fun & tolower(df$res) %in% show, , drop = FALSE]
      if (nrow(sub) == 0) next
      cli::cli_h3("Function {.strong {.code {fun}}}:")
      msg.size <- max(nchar(sub$msg), 30)
      cat(sprintf(" %*s  %-*s | %s\n",
                  res.size, tocolour(sub$res), msg.size, sub$msg, sub$what),
          sep = "")
    }
  } else {
    max.name <- max(c(0, nchar(x$funs))) + 1
    for (ii in seq_along(x$runs)) {
      what <- attr(x$runs[[ii]], "what")
      sub <- cbind(x$runs[[ii]], fun = x$funs)
      sub <- sub[tolower(sub$res) %in% show, ]
      if (nrow(sub) == 0) next()
      cli::cli_h3("Test input [[{ii}]]: {.strong {what}}")
      cat(sprintf("%*s  %*s  %s\n",
                  max.name, sub$fun, res.size, tocolour(sub$res), sub$msg),
          sep = "")
    }
  }
  cat("\n", summary.stats, "\n")
}

#' Extract the results for a specific test input
#'
#' This extracts the raw results for a single test input by its position,
#' returning the corresponding data frame from the `$runs` list.
#'
#' @param x An object of class `cbtf`.
#' @param i An index between 1 and the number of test inputs used.
#'
#' @return
#' If the index is valid, a data frame containing the following columns and
#' attributes:
#' \item{res}{One of "OK", "FAIL", "WARN" or "SKIP" for each combination of
#'       function and input tested (see the *Value* section in [fuzz]).}
#' \item{msg}{The message received in case of error, warning or skip,
#'       or an empty string if no failure occurred.}
#' \item{attr(*, "what")}{The character representation of the input tested.}
#' Otherwise, `FALSE`.
#'
#' @examples
#' res <- fuzz(funs = c("list", "matrix", "mean"),
#'             what = test_inputs(c("numeric", "raw")))
#' res[[6]]
#'
#' @export
`[[.cbtf` <- function(x, i) {
  i %in% seq_along(x$runs) && return(x$runs[[i]])
}

#' Compute the number of tests performed
#'
#' This returns the total number of tests as the product of the number of
#' inputs used and the number of functions tested.
#'
#' @param x An object of class `cbtf`.
#'
#' @return
#' An integer corresponding to the number of tests performed in a run.
#'
#' @examples
#' res <- fuzz(funs = c("list", "matrix", "mean"),
#'             what = test_inputs(c("numeric", "raw")))
#' length(res)
#'
#' @export
length.cbtf <- function(x) {
  length(x$runs) * length(x$runs[[1]]$res)
}

##===========================================================================
##
## Copyright (c) 2025-2026 Marco Colombo
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.
##
##===========================================================================

#' @title Validate that an argument is of the specified class
#'
#' @param arg Argument to validate.
#' @param class A character string for the candidate class or type.
#' @param null.ok Whether a `NULL` value should be considered valid (`FALSE`
#'        by default).
#' @param from Name of the caller function.
#' @param scalar Whether to consider the argument valid only if it's a scalar
#'        value (`FALSE` by default).
#' @param min Minimum value considered, or `NULL`.
#' @param choices Vector of accepted values, or `NULL`.
#' @param remove_empty Whether zero-length strings should be discarded from
#'        `arg` before checking if it is empty (`FALSE` by default).
#'
#' @return
#' Nothing in case of success, otherwise an error is thrown.
#'
#' @noRd
validate_class <- function(arg, class, null.ok = FALSE, from = "fuzz",
                           scalar = FALSE, min = NULL, choices = NULL,
                           remove_empty = FALSE) {
  !missing(arg) && is.null(arg) && null.ok && return()
  name <- sprintf("'%s'", all.vars(match.call())[1])
  if (missing(arg) || sum(inherits(arg, class)) == 0L ||
      (!is.list(arg) && length(arg) == 1 && (is.na(arg) || is.infinite(arg)))) {
    fuzz_error(name, "should be of class", paste(class, collapse = ", "),
               from = from)
  }
  scalar && length(arg) > 1 &&
    fuzz_error(name, "should be a single", class(arg), "value", from = from)
  if (!is.null(min) && (length(arg) == 0 || arg < min))
    fuzz_error(name, "should be at least", min, from = from)
  !is.null(choices) && !arg %in% choices &&
    fuzz_error(name, "should be one of", toString(sQuote(choices)))
  if (remove_empty)
    arg <- arg[nchar(arg) > 0]
  length(arg) == 0 &&
    fuzz_error(name, "is an empty", class(arg), from = from)
}

#' @title Stop with an error message
#'
#' @param ... Strings that are joined together in the error message.
#' @param from Name of the caller function.
#'
#' @noRd
fuzz_error <- function(..., from = "fuzz") {
  stop(do.call(paste, c(sprintf("[%s]", from), list(...))), call. = FALSE)
}

#' @title Check that a function can be fuzzed
#'
#' @param fun Name of the function to validate.
#' @param pkg Name of the package where functions are searched. A `NULL`
#'        value corresponds to the global namespace.
#' @param ignore_deprecated Whether deprecated functions should be ignored
#'        (`TRUE` by default).
#' @param num_args Number of arguments being fuzzed (1 by default).
#'
#' @return
#' In case of failure, a character string containing the reason why the
#' function cannot be fuzzed; otherwise the function itself.
#'
#' @noRd
check_fuzzable <- function(fun, pkg, ignore_deprecated = TRUE, num_args = 1L) {
  ## attempt to get a function from its name
  fun <- try(if (is.null(pkg)) get(fun)
             else utils::getFromNamespace(fun, pkg),
             silent = TRUE)

  ## skip non-existing names
  inherits(fun, "try-error") &&
    return(sprintf("Object not found in the %s namespace",
                   if (is.null(pkg)) "global" else sprintf("'%s'", pkg)))

  ## skip non-functions
  is.function(fun) ||
    return("Not a function")

  ## skip primitive functions that don't report arguments
  is.null(args(fun)) &&
    return("Doesn't specify number of arguments")

  ## skip functions accepting no arguments
  formal.args <- formals(args(fun))
  num.formals <- length(formal.args)
  num.formals == 0L &&
    return("Doesn't accept arguments")

  ## adjust the number of arguments
  if (any(grepl("...", names(formal.args), fixed = TRUE)))
    num.formals <- Inf

  ## skip functions accepting fewer arguments than provided
  num.formals < num_args &&
    return(sprintf("Accepts only up to %d arguments", num.formals))

  ## skip deprecated functions
  ignore_deprecated && any(grepl("\\.Deprecated", body(fun))) &&
    return("Deprecated function")

  fun
}

#' Generate coloured summary statistics from the results
#'
#' This computes summary statistics from the fuzzing results, prints a
#' message for the overall success or failure, and returns a summary string.
#'
#' @param object An object of class `cbtf`.
#' @param verbose Whether a message on the overall pass or fail of the fuzz
#'        run should be printed out (`TRUE` by default).
#'
#' @return
#' A summary results string formatted with ANSI colour codes.
#'
#' @noRd
compute_summary_stats <- function(object, verbose = TRUE) {
  results <- unlist(lapply(object$runs, function(x) x$res))
  success <- sum(results %in% c("FAIL", "WARN")) == 0
  if (verbose) {
    use.utf8 <- cli::is_utf8_output()
    symb.succ <- if (use.utf8) "\U1F3C3" else " "
    symb.fail <- if (use.utf8) "\U1F6A8" else " "
    if (success)
      cli::cli_alert_success(" {symb.succ} You didn't get caught by the fuzz!")
    else
      cli::cli_alert_danger(" {symb.fail}   CAUGHT BY THE FUZZ!   {symb.fail}")
  }

  stats <- as.list(table(results))
  summary.stats <- paste(tocolour("FAIL", sum(stats$FAIL)),
                         tocolour("WARN", sum(stats$WARN)),
                         tocolour("SKIP", sum(stats$SKIP)),
                         tocolour("OK",   sum(stats$OK), success),
                         sep = " | ")
  paste("[", summary.stats, "]")
}

#' Add colour formatting to a string
#'
#' @param res A character vector with "OK", "SKIP", "WARN" or "FAIL".
#' @param num A numerical value: if positive and `colour` is `TRUE`, then the
#'        `res` label is coloured; if finite, the `num` is appended to the
#'        (coloured) label.
#' @param colour A logical value that determines if the colour should be
#'        applied. If `FALSE`, nothing gets coloured independently of `num`.
#'
#' @return
#' A character vector formatted with ANSI colour codes.
#'
#' @noRd
tocolour <- function(res, num = Inf, colour = TRUE) {
  if (num > 0 && colour) {
    cols <- list(FAIL = cli::col_yellow("FAIL"),
                 WARN = cli::col_magenta("WARN"),
                 SKIP = cli::col_blue("SKIP"),
                 OK   = cli::col_green("OK"))
    res <- mapply(function(x) cols[[x]], res)
  }
  paste0(res, if (!is.infinite(num)) sprintf(" %d", num))
}

#' Append to each input a listified call to that input
#'
#' @param input A named list.
#'
#' @return
#' A named list with the original elements followed by their listified
#' version.
#'
#' @noRd
append_listified <- function(input) {
  transformed <- lapply(input, list)
  if (any(nzchar(names(transformed))))
    names(transformed) <- sprintf("list(%s)", names(transformed))
  c(input, transformed)
}

#' Retrieve or generate names for all elements in a list
#'
#' @param input A quosure of a list of possibly named elements.
#' @param use_names Whether the original element names should be used instead
#'        of the deparsed inputs when present.
#'
#' @return
#' A character vector of names.
#'
#' @noRd
get_element_names <- function(input, use_names) {
  if (rlang::quo_is_call(input, "list")) {
    ## extract the call expression from a literal list(...) supplied as a quosure
    expr <- rlang::quo_get_expr(input)
  } else {
    ## the quosure contains a symbol bound to a list: evaluate it to get the
    ## actual list value, prepending the NA placeholder so that further down
    ## as.list(expr)[-1] yields the list elements consistently
    expr <- c(NA, rlang::eval_tidy(input))
  }

  ## get the list elements and discard the list head
  elems <- as.list(expr)[-1]

  ## deparse each element to produce default names
  nm <- vapply(elems, deparse, character(1), nlines = 1)

  ## use the provided names for named elements
  if (use_names) {
    idx.named <- nzchar(names(elems))
    nm[idx.named] <- names(elems)[idx.named]
  }
  unname(nm)
}

#' Create a list of argument lists modified according to the given inputs
#'
#' This function generates all inputs to be tested. It is a precondition that
#' `args` and `what` cannot be both `NULL`.
#'
#' @param what A named list of inputs.
#' @param args A named list of arguments.
#' @param keys A vector of argument keys (original argument names, `NULL` by
#'        default). Arguments with non-empty names starting with `..` are
#'        treated as fixed (not fuzzed).
#'
#' @return
#' A list of named lists with the given arguments in turn replaced by each of
#' the inputs. Each of the generated argument lists is named by concatenating
#' the (possibly keyed) deparsed representation of its elements. If `keys` is
#' not empty, it is used to assign names to the individual elements in the
#' inner lists. Argument lists with the same names are removed.
#'
#' @noRd
modify_args <- function(what, args, keys = NULL) {
  ## a list label is a comma-separated concatenation of element names, with
  ## element names transformed into the form `key =` if a key was given
  list_label <- function(vals, idx = NULL, what_name = NULL) {
    vals[idx] <- what_name
    vals[is.key] <- paste(keys[is.key], "=", vals[is.key])
    paste(vals, collapse = ", ")
  }

  ## no arguments provided: return each element of `what` wrapped in a list
  is.null(args) && return(lapply(what, list))

  ## display labels for the argument values (pre-set in fuzz() from the
  ## quoted expression); original argument names are in `keys`
  vals_names <- names(args)
  is.key <- nzchar(keys)

  ## mark the arguments that should not be fuzzed
  idx.fixed <- which(startsWith(keys %||% "", ".."))
  if (!is.null(keys)) {
    keys <- gsub("^\\.\\.", "", keys)
    if (length(idx.fixed) == length(args) && !is.null(what)) {
      cli::cli_alert_warning(c("'args' contains only fixed elements, ",
                               "'what' will be ignored"))
      what <- NULL
    }
  }

  ## `args` must be named with its (cleaned) keys, as do.call() uses them
  ## to pass arguments to the function being called
  names(args) <- keys

  ## no inputs provided: return `args` without changes wrapped in a named list
  is.null(what) && return(stats::setNames(list(args), list_label(vals_names)))

  ## create a list of modified argument lists, where the element at index
  ## idx is replaced in turn by each element in `what`
  res <- unlist(lapply(seq_along(args), function(idx) {
    idx %in% idx.fixed && return(NULL)
    new <- lapply(what, function(inp) {
      new <- args
      new[idx] <- list(inp)
      new
    })
    labels <- vapply(names(what), function(what_name) {
      list_label(vals_names, idx, what_name)
    }, character(1))
    stats::setNames(new, labels)
  }), recursive = FALSE)

  ## keep unique elements according to their names
  res[!duplicated(names(res))]
}

test_that("validate_class", {
  testthat::skip_on_cran()

  arg <- data.frame()
  expect_error(validate_class(arg, "character"),
               "[fuzz] 'arg' should be of class character",
               fixed = TRUE)
  expect_error(validate_class(arg, "character", from = "function_name"),
               "[function_name] 'arg' should be of class character",
               fixed = TRUE)
  expect_error(validate_class(arg, "data.frame"),
               "'arg' is an empty data.frame")
  expect_error(validate_class(iris, "data.frame", scalar = TRUE),
               "'iris' should be a single data.frame value")
  expect_error(validate_class(letters, "character", scalar = TRUE),
               "'letters' should be a single character value")
  expect_error(validate_class(letters, c("numeric", "character"), scalar = TRUE),
               "^\\[fuzz\\] 'letters' should be a single character value$")
  arg <- "Inf"
  expect_error(validate_class(arg, "numeric"),
               "'arg' should be of class numeric")
  arg <- NULL
  expect_error(validate_class(arg, "character"),
               "'arg' should be of class character")
  arg <- ""
  expect_error(validate_class(arg, "character", remove_empty = TRUE),
               "'arg' is an empty character")
  arg <- list()
  expect_error(validate_class(arg, "list", remove_empty = TRUE),
               "'arg' is an empty list")
  arg <- list("")
  expect_error(validate_class(arg, "list", remove_empty = TRUE),
               "'arg' is an empty list")
  arg <- 1
  expect_error(validate_class(arg, "numeric", min = 2),
               "'arg' should be at least 2")
  arg <- numeric()
  expect_error(validate_class(arg, "numeric", min = 2),
               "'arg' should be at least 2")
  arg <- "error"
  expect_error(validate_class(arg, "character", choices = c("c1", "c2")),
               "'arg' should be one of 'c1', 'c2'")

  expect_silent(validate_class(iris, "data.frame"))
  expect_silent(validate_class(NULL, "data.frame", null.ok = TRUE))
  expect_silent(validate_class(iris, "data.frame", remove_empty = TRUE))
  expect_silent(validate_class("", "character"))
  expect_silent(validate_class(list(""), "list"))
  expect_silent(validate_class(2, "numeric", min = 1))
  expect_silent(validate_class("c1", "character", choices = c("c1", "c2")))
})

test_that("fuzz_error", {
  testthat::skip_on_cran()

  expect_error(fuzz_error("message"),
               "[fuzz] message", fixed = TRUE)
  expect_error(fuzz_error("part 1", "part 2", from = "function_name"),
               "[function_name] part 1 part 2", fixed = TRUE)
})

test_that("check_fuzzable", {
  testthat::skip_on_cran()

  expect_equal(check_fuzzable(".not.existing.", NULL),
               "Object not found in the global namespace")
  expect_equal(check_fuzzable(".not.existing.", "package_name"),
               "Object not found in the 'package_name' namespace")
  expect_equal(check_fuzzable(".Device", NULL),
               "Not a function")
  expect_equal(check_fuzzable(":", NULL),
               "Doesn't specify number of arguments")
  expect_equal(check_fuzzable("Sys.Date", NULL),
               "Doesn't accept arguments")
  expect_equal(check_fuzzable("+", NULL, num_args = 3),
               "Accepts only up to 2 arguments")

  ## check that the ... is interpreted correctly
  expect_equal(check_fuzzable("list", NULL, num_args = 3),
               list)

  ## must use `assign` otherwise the name cannot be found by the `get` call
  assign(".with.readline.", function(val) readline("Test"), envir = .GlobalEnv)
  expect_equal(check_fuzzable(".with.readline.", NULL),
               .with.readline.)
  rm(".with.readline.", envir = .GlobalEnv)

  expect_true(is.function(check_fuzzable("list", NULL)))
  expect_true(is.function(check_fuzzable("mean", NULL)))

  assign(".deprecated.", envir = .GlobalEnv,
         function(arg) .Deprecated("new"))
  expect_true(is.function(check_fuzzable(".deprecated.", NULL,
                                         ignore_deprecated = FALSE)))
  expect_equal(check_fuzzable(".deprecated.", NULL, ignore_deprecated = TRUE),
               "Deprecated function")
  rm(".deprecated.", envir = .GlobalEnv)
})

test_that("tocolour", {
  testthat::skip_on_cran()

  expect_equal(tocolour("FAIL"),
               as.character(cli::col_yellow("FAIL")))
  expect_equal(tocolour("WARN"),
               as.character(cli::col_magenta("WARN")))
  expect_equal(tocolour("SKIP"),
               as.character(cli::col_blue("SKIP")))
  expect_equal(tocolour("OK"),
               as.character(cli::col_green("OK")))
  expect_equal(tocolour("FAIL", 0),
               "FAIL 0")
  expect_equal(tocolour("FAIL", 1),
               paste(cli::col_yellow("FAIL"), 1))
  expect_equal(tocolour("FAIL", 1, FALSE),
               "FAIL 1")
})

test_that("namify", {
  testthat::skip_on_cran()

  expect_equal(namify(NA),
               list("NA" = NA))
  expect_equal(namify(12, one = 1, "", iris, matrix(1, 0, 1)),
               list("12" = 12,
                    "one" = 1,
                    "\"\"" = "",
                    "iris" = iris,
                    "matrix(1, 0, 1)" = matrix(1, 0, 1)))
})

test_that("append_listified", {
  testthat::skip_on_cran()

  ## list
  expect_equal(append_listified(list(NA)),
               list(NA, list(NA)))
  expect_equal(append_listified(list(list())),
               list(list(), list(list())))
  res <- append_listified(list(matrix = matrix()))
  expect_equal(res,
               list(matrix = matrix(), "list(matrix)" = list(matrix())))
  res <- append_listified(list(`NULL` = NULL, `NA` = NA, `0L` = 0L))
  expect_named(res,
               c("NULL", "NA", "0L", "list(NULL)", "list(NA)", "list(0L)"))
  res <- append_listified(list(m0 = matrix(0, 0, 0), m1 = matrix(1, 0, 1)))
  expect_named(res,
               c("m0", "m1", "list(m0)", "list(m1)"))

  ## alist
  expect_equal(append_listified(alist(NA)),
               list(NA, list(NA)))
  expect_equal(append_listified(alist(list())),
               list(quote(list()), list(quote(list()))))
})

test_that("get_element_names", {
  testthat::skip_on_cran()

  expect_equal(get_element_names(rlang::quo(list(NA, NULL)),
                                 use_names = TRUE),
               c("NA", "NULL"))
  expect_equal(get_element_names(rlang::quo(list(first = NA, NULL)),
                                 use_names = TRUE),
               c("first", "NULL"))
  expect_equal(get_element_names(rlang::quo(list(first = NA, iris)),
                                 use_names = TRUE),
               c("first", "iris"))
  expect_equal(get_element_names(rlang::quo(list(df = iris, iris)),
                                 use_names = TRUE),
               c("df", "iris"))

  ## list passed indirectly
  what <- list(iris, data.frame())
  expect_equal(get_element_names(rlang::quo(what),
                                 use_names = TRUE),
               c("structure(list(Sepal.Length = c(5.1, 4.9, 4.7, 4.6, 5, 5.4, 4.6, ",
                 "structure(list(), names = character(0), row.names = integer(0), class = \"data.frame\")"))

  what <- list(first = iris, data.frame())
  expect_equal(get_element_names(rlang::quo(what),
                                 use_names = TRUE),
               c("first",
                 "structure(list(), names = character(0), row.names = integer(0), class = \"data.frame\")"))
  expect_equal(get_element_names(rlang::quo(test_inputs("na")),
                                 use_names = TRUE),
               c("NA_real_", "NA_integer_", "NA_character_", "NA"))
})

test_that("modify_args", {
  testthat::skip_on_cran()

  expect_equal(modify_args(namify(1, 2, 3), args = NULL),
               lapply(namify(1, 2, 3), list))
  expect_equal(modify_args(namify(NA, 1.2), args = namify(TRUE, 2:5)),
               structure(list(list(NA, 2:5), list(1.2, 2:5),
                              list(TRUE, NA), list(TRUE, 1.2)),
                         names = c("NA, 2:5", "1.2, 2:5",
                                   "TRUE, NA", "TRUE, 1.2")))

  ## unique
  expect_equal(modify_args(namify(1, 2, 3), args = namify(1, 2)),
               structure(list(list(1, 2), list(2, 2), list(3, 2),
                              list(1, 1), list(1, 3)),
                         names = c("1, 2", "2, 2", "3, 2",
                                   "1, 1", "1, 3")))

  ## NULL
  expect_equal(modify_args(namify(0, NULL), args = namify(1, 2)),
               structure(list(list(0, 2), list(NULL, 2),
                              list(1, 0), list(1, NULL)),
                         names = c("0, 2", "NULL, 2",
                                   "1, 0", "1, NULL")))

  ## keys
  res <- modify_args(what = NULL, args = namify(1, 2, 3),
                     keys = NULL)
  expect_equal(res,
               structure(list(list(1, 2, 3)),
                         names = "1, 2, 3"))
  expect_named(res[[1]], NULL)

  res <- modify_args(what = NULL, args = namify(1, 2, 3),
                     keys = c("x", "y", "z"))
  expect_equal(res,
               structure(list(list(x = 1, y = 2, z = 3)),
                         names = "x = 1, y = 2, z = 3"))
  expect_named(res[[1]], c("x", "y", "z"))

  res <- modify_args(what = NULL, args = namify(iris, TRUE),
                     keys = c("", "verbose"))
  expect_equal(res,
               structure(list(list(iris, verbose = TRUE)),
                         names = c("iris, verbose = TRUE")))

  res <- modify_args(what = namify(0, NULL), args = namify(1, 2),
                     keys = c("x", ""))
  expect_equal(res,
               structure(list(list(x = 0, 2), list(x = NULL, 2),
                              list(x = 1, 0), list(x = 1, NULL)),
                         names = c("x = 0, 2", "x = NULL, 2",
                                   "x = 1, 0", "x = 1, NULL")))
  expect_named(res[[1]], c("x", ""))

  ## fixed keys
  res <- modify_args(what = namify(NA, NULL), args = namify(1, 2),
                     keys = c("", "..x"))
  expect_equal(res,
               structure(list(list(NA, x = 2), list(NULL, x = 2)),
                         names = c("NA, x = 2", "NULL, x = 2")))
  expect_message(res <- modify_args(what = namify(NA, NULL), args = namify(1, 2),
                                    keys = c("..x", "..y")),
                 "'args' contains only fixed elements")
  expect_equal(res,
               structure(list(list(x = 1, y = 2)),
                         names = c("x = 1, y = 2")))
  expect_no_message(modify_args(what = NULL, args = namify(1, 2),
                                keys = c("..x", "..y")),
                    message = "'args' contains only fixed elements")
})

##===========================================================================
##
## Copyright (c) 2025-2026 Marco Colombo
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.
##
##===========================================================================

setup_queue <- function(funs, what, timeout,
                        package, ignore_patterns, ignore_warnings) {
  n_tasks <- length(funs) * length(what)
  daemons <- mirai::info()[["connections"]]
  results <- vector(mode = "list", length = n_tasks)
  running <- list()
  current <- 0L
  finishd <- 0L
  message <- ""
  char <- names(what)

  ## fuzzing engine
  fuzzer <- quote({
    fun <- check_fuzzable(name, package, ignore_deprecated = FALSE, length(what))
    is.character(fun) && return(data.frame(res = "SKIP", msg = fun))

    whitelist_and_label <- function(label, msg) {
      res <- if ((label == "WARN" && ignore_warnings) ||
                 is.null(msg) ||
                 all(grepl(name, msg, fixed = TRUE)) ||
                 all(grepl(ignore_patterns, msg))) {
               "OK"
             } else {
               label
             }
      data.frame(res = res, msg = paste(msg, collapse = " | "))
    }

    warnings <- NULL
    tryCatch(
        withCallingHandlers({
          do.call(fun, what)
          whitelist_and_label("WARN", warnings)
        }, warning = function(w) {
          warnings <<- c(warnings, trimws(conditionMessage(w)))
        }),
        error = function(e) {
          whitelist_and_label("FAIL", conditionMessage(e)[1])
        })
  })

  ## export common data to the daemons
  ## for performance reasons, we pass only the functions we need
  env <- sapply(funs, function(x) .GlobalEnv[[x]])
  env[vapply(env, is.null, logical(1))] <- NULL
  env <- as.environment(env)
  list2env(list(package = package,
                ignore_patterns = ignore_patterns,
                ignore_warnings = ignore_warnings,
                check_fuzzable = check_fuzzable,
                fuzzer = fuzzer), envir = env)
  mirai::everywhere({}, env)

  ## Run tasks and collect the results
  process <- function() {
    cli::cli_progress_bar(format = paste(
                              "{cli::pb_spin} Test input: {cli::pb_bar}|",
                              "{.timestamp {cli::pb_current}/{cli::pb_total}}",
                              " {.strong {message}}"),
                          auto_terminate = FALSE,
                          total = length(what))

    ## process all the tasks
    tic <- Sys.time()
    while (run_tasks()) {
      cli::cli_progress_update(set = finishd)
    }
    elapsed <- difftime(Sys.time(), tic, units = "secs")
    elapsed <- if (elapsed < 1) paste0(round(elapsed * 1000), "ms")
               else paste0(round(elapsed, 1), "s")
    cli::cli_progress_done()
    cli::cli_alert_info("{n_tasks} tests run  {.timestamp {elapsed}}")

    collect_results()
  }

  ## Start running a task
  run_tasks <- function() {
    while (length(running) < daemons && current < n_tasks) {
      fidx <- current %% length(funs) + 1L
      widx <- current %/% length(funs) + 1L
      args <- list(name = funs[[fidx]],
                   what = what[[widx]],
                   char = char[[widx]])
      running[[length(running) + 1]] <<- list(index = current,
                                              mirai = mirai::mirai(eval(fuzzer),
                                                                   .args = args,
                                                                   .timeout = timeout * 1000))
      current <<- current + 1L
    }

    get_tasks()
  }

  ## Get the results from a task
  get_tasks <- function() {
    ## return early if no more tasks are running
    length(running) || return(current < n_tasks)

    ## wait until the first task completion
    mirai::race_mirai(lapply(running, `[[`, "mirai"))

    ## collect the completed runs
    rm.idx <- NULL
    lapply(seq_along(running), function(idx) {
      !mirai::unresolved(running[[idx]]$mirai) || return()
      res <- running[[idx]]$mirai$data
      if (mirai::is_error_value(res) && as.integer(res) == 5L) {
        res <- data.frame(res = "OK",
                          msg = sprintf("Timed out after %d seconds", timeout))
      }

      results[[running[[idx]]$index + 1L]] <<- res
      rm.idx <<- c(rm.idx, idx)
    })

    running[rm.idx] <<- NULL
    if (length(running)) {
      finishd <<- min(sapply(running, `[[`, "index")) %/% length(funs) + 1L
      message <<- char[[finishd]]
    }

    length(running) || current < n_tasks
  }

  ## Collect the final results
  collect_results <- function() {
    res <- do.call(rbind, results)

    ## group the results by input
    lapply(seq_along(what), function(idx) {
      sub <- res[(idx - 1) * length(funs) + seq_along(funs), ]
      rownames(sub) <- NULL
      attr(sub, "what") <- char[[idx]]
      sub
    })
  }

  environment()
}

##=============================================================================
##
## Copyright (c) 2025-2026 Marco Colombo
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.
##
##=============================================================================


#' CBTF: Caught by the Fuzz! A minimalistic fuzz-test runner
#'
#' This package implements a very simple mechanism for fuzz-testing functions
#' in the public interface of an R package.
#'
#' Fuzz testing helps identify functions that lack sufficient argument validation,
#' and uncovers sets of inputs that, while valid by function signature, may
#' cause issues within the function body.
#'
#' The core functionality of the package is in [fuzz], which calls each
#' provided function with a certain input and records the output produced.
#' If an error or a warning is generated, this is captured and reported to the
#' user, unless it matches a pattern of whitelisted messages. The objects
#' returned by [fuzz] can be inspected with [summary.cbtf] and [print.cbtf].
#'
#' Whitelisting can also be done after a fuzz run has been completed via the
#' [whitelist] function, so that only messages that need to be acted upon are
#' actually shown. Using [whitelist] has the advantage of not requiring the
#' completion of a fuzz run of all functions over all inputs again.
#'
#' Note that [fuzz] relies on the `mirai` package for asynchronous operations
#' and parallelisation, and execution occurs on persistent background processes.
#' These can be started automatically by specifying the `daemons` option;
#' alternatively, they can be set up manually with the `mirai::daemons()`
#' function; refer to the original `mirai` documentation for a complete
#' description of its arguments and behaviour.
#'
#' The helper function [get_exported_functions] identifies the functions
#' in the public interface of a given package, facilitating the generation of
#' the list of functions to be fuzzed.
#'
#' The helper function [test_inputs] is invoked by [fuzz] if the user doesn't
#' specify the set of inputs to be tested. By default it generates a large set
#' of potentially problematic inputs, but these can be limited just to the
#' desired classes of inputs.
#'
"_PACKAGE"

##===========================================================================
##
## Copyright (c) 2024-2026 Marco Colombo
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.
##
##===========================================================================

#' Get the names of the exported functions of a package
#'
#' This function extracts the exports from the namespace of the given package
#' via [getNamespaceExports] and discards non-fuzzable objects (non-functions
#' and functions with no arguments). The set of names returned can be further
#' restricted via the `ignore_names` and `ignore_deprecated` arguments.
#'
#' @param package Name of the package where functions are searched.
#' @param ignore_names Names of functions to ignore: these are removed from
#'        the names returned. This can be helpful, for example, to discard
#'        function aliases.
#' @param ignore_deprecated Whether deprecated functions should be ignored
#'        (`TRUE` by default).
#'
#' @return
#' A character vector of the names of the fuzzable functions exported from
#' the given package, with the `"package"` attribute set. This can be used
#' directly as the `funs` argument of [fuzz] without the need to specify the
#' `package` argument.
#'
#' @examples
#' ## get the fuzzable functions in the public interface of this package
#' funs <- get_exported_functions("CBTF")
#'
#' @seealso [fuzz]
#'
#' @export
get_exported_functions <- function(package, ignore_names = "",
                                   ignore_deprecated = TRUE) {
  from <- "get_exported_functions"
  validate_class(package, "character", from = from,
                 scalar = TRUE, remove_empty = TRUE)
  validate_class(ignore_names, "character", from = from)
  validate_class(ignore_deprecated, "logical", from = from, scalar = TRUE)
  funs <- tryCatch(sort(getNamespaceExports(package)),
                   error = function(e) fuzz_error(e$message, from = from))

  ## keep only fuzzable functions
  keep.idx <- sapply(funs, function(x) {
    is.function(check_fuzzable(x, package, ignore_deprecated))
  })
  funs <- setdiff(funs[keep.idx], ignore_names)
  attr(funs, "package") <- package
  funs
}

#' Fuzz-test the specified functions
#'
#' The fuzzer calls each function in `funs` with the argument list provided in
#' `args` (each of its elements in turn modified by each object in `what`) and
#' records any errors or warnings that are thrown. If no error occurs within
#' the first `timeout` seconds, the execution of the function being fuzzed is
#' interrupted and the next one is started.
#'
#' @details
#' ## Multiple arguments
#'
#' A list of arguments to be passed to the functions being fuzzed can be
#' provided via the `args` argument. Each element in that list is modified in
#' turn by each object in `what` and the resulting list of arguments is then
#' passed to each function via `do.call()`. If more arguments are given than
#' the number of formal arguments accepted by a function, that function will
#' produce a "SKIP" result.
#'
#' If arguments are named, they will be passed with their names to the fuzzed
#' functions. If a function doesn't have a formal argument of that name and
#' doesn't accept `...`, then the standard R behaviour is to return an "unused
#' argument" error. This is whitelisted by default in `fuzz()`, and the
#' corresponding result status is set to "OK".
#'
#' It is possible to define arguments that should remain unchanged while
#' fuzzing by prefixing their names with `..`. These arguments will use the
#' values assigned in `args` without modification. For example, to ensure that
#' the argument `na.rm` is always set to `TRUE`, it should be specified as
#' `..na.rm = TRUE` in `args`. If all elements in `args` are fixed, `what` is
#' ignored and all functions in `funs` will be called with the provided `args`
#' list.
#'
#' ## Parallel execution
#'
#' The implementation uses `mirai` as a backend to execute tasks asynchronously
#' in parallel worker processes. The function can start a pool of persistent
#' background processes (daemons) of size given by the `daemons` argument
#' (note that starting more daemons than available cores yields no benefit).
#' Alternatively, the function can also make use of already active daemons
#' started with the [mirai::daemons] function: this allows to control in
#' greater detail the number of processes to use, which can also be remote.
#'
#' ## Whitelisting
#'
#' In order to reduce the number of false positive results produced, this
#' function applies the following set of rules, to establish if an error or
#' warning condition should be ignored (whitelisting):
#'
#' * If the name of the function appears in the error or warning message, as
#'   it is considered that the condition has been handled by the developer.
#' * If the error or warning message contains the text "is missing, with no
#'   default", which is produced when a missing argument is used without a
#'   value being assigned to it.
#' * If the error or warning message contains any of the patterns specified
#'   in `ignore_patterns`. In case of multiple warnings being captured, the
#'   result is whitelisted only if _all_ warnings match any of the patterns.
#' * If a warning is thrown but `ignore_warnings = TRUE` is set.
#'
#' In all whitelisted cases, the result is "OK", and the message that
#' was received is stored in the `$msg` field (see the *Value* section).
#'
#' *Note:* Whitelisting can also be applied post-hoc on the results of a fuzz
#' run using the [whitelist] function.
#'
#' @param funs A character vector of function names to test. If the vector has
#'        a `"package"` attribute and no `package` argument is given, functions
#'        are loaded from the namespace specified in that attribute.
#' @param what A list of objects; each is used, in turn, to modify the list of
#'        arguments in `args` before calling each of the functions in `funs`.
#'        If no inputs are provided, a default set of inputs generated by
#'        [test_inputs] will be used. If set to `NULL`, then `args` must be
#'        specified, and all functions will be called with that exact list of
#'        arguments with no fuzzing occurring.
#' @param args A list of arguments to pass to the functions being fuzzed. Each
#'        element in the list is in turn replaced by each object in `what`,
#'        then each modified argument list is used to fuzz the functions in
#'        `funs`. Argument names are preserved and used as named arguments in
#'        the fuzzed functions. If `NULL` (default), only the first argument of
#'        each function is fuzzed.
#' @param package Name of the package where functions are searched. If `NULL`
#'        (default), the function first checks the `"package"` attribute of
#'        `funs`, and if that is not set, names are searched in the global
#'        namespace.
#' @param listify_what Whether each input in `what` should also be tested
#'        in its listified version (`FALSE` by default). When set to `TRUE`,
#'        if `what` is `list(x = x)`, the function will operate as if it
#'        were `list(x = x, "list(x)" = list(x))`, for any input object `x`.
#' @param ignore_patterns One or more strings containing regular expressions
#'        to match the errors to ignore. The strings "unused argument" and
#'        "is missing, with no default" are always ignored.
#' @param ignore_warnings Whether warnings should be ignored (`FALSE` by
#'        default).
#' @param daemons Number of daemons to use (2 by default). As many `mirai`
#'        daemons as specified will be started when entering the function and
#'        closed at the end, unless active daemons are already available, in
#'        which case the argument is ignored and the active daemons are used.
#' @param timeout Number of seconds (2 by default) after which the function
#'        being fuzzed is interrupted with result status set to "OK".
#'
#' @return
#' An object of class `cbtf` that stores the results obtained for each of the
#' functions tested. This contains the following fields:
#' \item{runs}{a list of data frames, each containing the results of fuzzing
#'       all the functions in `funs` with one of the inputs in `what`. The
#'       data frame contains the following columns and attributes:\cr
#'       - `res`: The result of the fuzz test, see below for the possible
#'         values.\cr
#'       - `msg`: The error or warning message returned by the function, if
#'          any.\cr
#'       - `attr(*, "what")`: The character representation of the input
#'         tested.
#' }
#' \item{funs}{a vector of names of the functions tested.}
#' \item{args}{a named list of arguments, with names generated by deparsing
#'       the `args` argument if not already provided.}
#' \item{package}{a character string specifying the package name where
#'       function names were searched, or `NA` if none was provided.}
#' \item{ignore_patterns}{The value of the `ignore_patterns` argument.}
#' \item{ignore_warnings}{The value of the `ignore_warnings` argument.}
#'
#' The `res` column in each of the data frames in the `$runs` field can
#' contain the following values:
#' * **OK**: either no error or warning was produced (in which case, the `msg`
#'   entry is left blank), or it was whitelisted (in which case, the message
#'   received is stored in `msg`), or it was timed out (in which case, `msg`
#'   records that a timeout was applied).
#' * **SKIP**: no test was run, either because the given name cannot be found, or
#'   it doesn't correspond to a function, or the function accepts no arguments,
#'   or more arguments were provided than the function accepts;
#'   the exact reason is given in `msg`.
#' * **WARN**: a warning was thrown for which no whitelisting occurred and
#'   `ignore_warnings = FALSE`; its message is stored in `msg`.
#' * **FAIL**: an error was thrown for which no whitelisting occurred; its message
#'   is stored in `msg`.
#'
#' @examples
#' ## set up persistent background processes
#' mirai::daemons(2L)
#'
#' ## this should produce no errors
#' res <- fuzz(funs = c("list", "matrix", "mean"),
#'             what = test_inputs(c("numeric", "raw")))
#' summary(res)
#'
#' ## display all results even for successful tests
#' print(res, show = "all")
#'
#' ## this will catch an error (false positive)
#' fuzz(funs = "matrix",  what = test_inputs("scalar"))
#'
#' ## apply a whitelist pattern to remove the false positive
#' fuzz(funs = "matrix",  what = test_inputs("scalar"),
#'      ignore_patterns = "'data' must be of a vector type")
#'
#' ## close the background processes
#' mirai::daemons(0L)
#'
#' @seealso [get_exported_functions], [test_inputs], [whitelist],
#' [summary.cbtf], [print.cbtf]
#'
#' @export
fuzz <- function(funs, what = test_inputs(), args = NULL,
                 package = NULL, listify_what = FALSE,
                 ignore_patterns = "", ignore_warnings = FALSE,
                 daemons = 2L, timeout = 2) {

  ## defuse the user inputs
  what.quo <- rlang::enquo(what)
  args.quo <- rlang::enquo(args)

  ## input validation
  validate_class(funs, "character", remove_empty = TRUE)
  validate_class(what, "list", null.ok = TRUE)
  validate_class(args, "list", null.ok = TRUE)
  if (is.null(what) && is.null(args))
    fuzz_error("'what' and 'args' cannot be both NULL")
  if (is.null(package)) {
    package <- attr(funs, "package")
  } else {
    validate_class(package, "character", scalar = TRUE, remove_empty = TRUE)
  }
  validate_class(listify_what, "logical", scalar = TRUE)
  validate_class(ignore_patterns, "character")
  validate_class(ignore_warnings, "logical", scalar = TRUE)
  validate_class(daemons, c("integer", "numeric"), scalar = TRUE, min = 1)
  validate_class(timeout, c("numeric", "integer"), scalar = TRUE, min = 1)

  ## cap the timeout so that converting to milliseconds won't overflow
  timeout <- min(timeout, 10000)

  ## start as many daemons as specified by the `daemons` argument, unless
  ## there are daemons already running
  mirai::daemons_set() || {
    on.exit(mirai::daemons(0L), add = TRUE)
    mirai::daemons(n = daemons)
  }

  ## preserve the original argument names, as names(args) will be overwritten
  ## by get_element_names() with deparsed display labels
  keys <- names(args)

  ## assign display labels to `what` and `args` from the quoted expressions
  if (!is.null(what))
    names(what) <- get_element_names(what.quo, use_names = TRUE)
  if (!is.null(args)) {
    names(args) <- get_element_names(args.quo, use_names = FALSE)
  }

  ## expand the set of inputs with their listified version
  if (listify_what)
    what <- append_listified(what)

  ## expand the set of inputs according to the arguments provided
  what <- modify_args(what, args, keys)

  ## join all regular expression patterns
  joined_patterns <- paste0(c(ignore_patterns,
                              "unused argument",
                              "is missing, with no default"),
                            collapse = "|")
  joined_patterns <- gsub("^\\|", "", joined_patterns) # remove extra |

  ## start fuzzing
  cli::cli_alert_info(c("Fuzzing {length(funs)} function{?s} ",
                        "with {length(what)} input{?s} ",
                        "(using {mirai::info()[['connections']]} daemon{?s})"))
  if (is.null(package))
    cli::cli_alert_info(c("Functions will be searched in the global namespace ",
                          "as {.arg package} was not specified"))

  ## ensure that we always show some progress
  if (is.null(getOption("cli.progress_show_after"))) {
    opt <- options(cli.progress_show_after = 0.1)
    on.exit(options(opt), add = TRUE)
  }

  ## create the queue
  queue <- setup_queue(funs, what, timeout = timeout,
                       package, joined_patterns, ignore_warnings)

  ## fuzz the functions asynchronously
  runs <- queue$process()

  ## returned object
  structure(list(runs = runs,
                 funs = funs,
                 package = package %||% NA,
                 ignore_patterns = ignore_patterns,
                 ignore_warnings = ignore_warnings),
            class = "cbtf")
}

#' Apply additional whitelist patterns to the results of a fuzz run
#'
#' This allows for post-hoc whitelisting of results according to the patterns
#' specified.
#'
#' @param object An object of class `cbtf`.
#' @param patterns One or more strings containing regular expressions to
#'        match the errors to whitelist.
#'
#' @return
#' An object of class `cbtf` with the additional whitelist patterns applied.
#'
#' @examples
#'
#' ## this reports a false positive result
#' (res <- fuzz(funs = "matrix", what = test_inputs("scalar")))
#'
#' ## with whitelisting, we can remove that
#' whitelist(res, "must be of a vector type")
#'
#' @seealso [fuzz]
#'
#' @export
whitelist <- function(object, patterns) {
  from <- "whitelist"
  validate_class(object, "cbtf", from = from)
  validate_class(patterns, "character", from = from, remove_empty = TRUE)

  ## join all regular expression patterns
  joined_patterns <- paste0(patterns, collapse = "|")
  joined_patterns <- gsub("^\\|", "", joined_patterns) # remove extra |

  ## apply the new whitelist patterns to errors and warnings
  object$runs <- lapply(object$runs, function(x) {
    x$res[grepl(joined_patterns, x$msg) & x$msg != "SKIP"] <- "OK"
    x
  })
  object$ignore_patterns <- setdiff(c(object$ignore_patterns, patterns), "")
  object
}

test_that("summary", {
  testthat::skip_on_cran()

  SW({
  res <- fuzz("list", list(NA, c(1, 2, 3)))
  sum <- summary(res)
  expect_s3_class(sum,
                  "data.frame")
  expect_named(sum,
               c("fun", "what", "res", "msg"))
  expect_equal(sum$what,
               c("NA", "c(1, 2, 3)"))
  })
  expect_s3_class(attr(sum, "summary_table"),
                  "table")

  expect_error(summary(res, tabulate = NA),
               "[summary] 'tabulate' should be of class logical",
               fixed = TRUE)
  expect_error(summary(res, tabulate = c(TRUE, FALSE)),
               "[summary] 'tabulate' should be a single logical value",
               fixed = TRUE)
  expect_snapshot(summary(res))
  expect_snapshot(summary(res, tabulate = FALSE))
})

test_that("print", {
  testthat::skip_on_cran()

  SW({
  expect_output(expect_pass_message(fuzz("list", list(NA, c(1, 2, 3)))),
                "OK 2")
  expect_output(expect_fail_message(fuzz("ls", list(NA, c(1, 2, 3)))),
                "FAIL 2")

  ## no failures
  res <- fuzz("list", list(NA, c(1, 2, 3)))
  expect_snapshot(print(res))
  expect_snapshot(print(res, group = "function", show = "all"))

  ## with failures and skips
  res <- fuzz(c("list", "median", "Sys.date"), list(NA), args = list(1:3, TRUE))

  expect_error(print(res, show = NA),
               "[print] 'show' should be of class character",
               fixed = TRUE)
  expect_snapshot(print(res))
  expect_snapshot(print(res, show = "all"))
  expect_snapshot(print(res, show = "skip"))
  expect_snapshot(print(res, show = "none"))
  expect_snapshot(print(res, group = "function", show = "all"))
  })
})

test_that("[[", {
  testthat::skip_on_cran()

  SW({
  res <- fuzz("list", list(NA, c(1, 2, 3)))
  expect_s3_class(res[[1]],
                  "data.frame")
  expect_equal(colnames(res[[1]]),
               c("res", "msg"))
  expect_equal(attr(res[[1]], "what"),
               "NA")
  expect_false(res[[0]])
  expect_false(res[[10]])
  })
})

## remove local binding generated further down with `assign()`
withr::defer(rm(".local_fun.", envir = .GlobalEnv))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(fuzz(NA, NULL),
               "'funs' should be of class character")
  expect_error(fuzz(NULL, NULL),
               "'funs' should be of class character")
  expect_error(fuzz(what = NA),
               "'funs' should be of class character")
  expect_error(fuzz(character(0), NULL),
               "'funs' is an empty character")
  expect_error(fuzz("list", list()),
               "'what' is an empty list")
  expect_error(fuzz("median", args = c(NA, TRUE)),
               "'args' should be of class list")
  expect_error(fuzz("median", what = NULL, args = NULL),
               "'what' and 'args' cannot be both NULL")
  expect_error(fuzz("list", package = letters),
               "'package' should be a single character value")
  expect_error(fuzz("list", package = ""),
               "'package' is an empty character")
  expect_error(fuzz("list", listify_what = NULL),
               "'listify_what' should be of class logical")
  expect_error(fuzz("list", listify_what = c(TRUE, FALSE)),
               "'listify_what' should be a single logical value")
  expect_error(fuzz("list", ignore_patterns = TRUE),
               "'ignore_patterns' should be of class character")
  expect_error(fuzz("list", list(NA), ignore_warnings = NA),
               "'ignore_warnings' should be of class logical")
  expect_error(fuzz("list", list(NA), ignore_warnings = c(TRUE, FALSE)),
               "'ignore_warnings' should be a single logical value")
  expect_error(fuzz("list", list(NA), daemons = NA_integer_),
               "'daemons' should be of class integer, numeric")
  expect_error(fuzz("list", list(NA), daemons = c(2, 3)),
               "'daemons' should be a single numeric value")
  expect_error(fuzz("list", list(NA), daemons = 0),
               "'daemons' should be at least 1")
  expect_error(fuzz("list", list(NA), timeout = NA_integer_),
               "'timeout' should be of class numeric, integer")
  expect_error(fuzz("list", list(NA), timeout = c(2, 3)),
               "'timeout' should be a single numeric value")
  expect_error(fuzz("list", list(NA), timeout = 0),
               "'timeout' should be at least 1")
})

test_that("check skipped functions", {
  testthat::skip_on_cran()

  SW({
  expect_skip_reason(fuzz("Sys.Date", list(NULL)),
                     "Doesn't accept arguments")
  expect_skip_reason(fuzz("iris", list(NULL)),
                     "Not a function")
  expect_skip_reason(fuzz(".not.found.", list(NULL)),
                     "Object not found in the global namespace")
  expect_skip_reason(fuzz(".not.found.", list(NULL), package = "CBTF"),
                     "Object not found in the 'CBTF' namespace")

  ## must use `assign` otherwise the name cannot be found by the `get` call
  assign(".local_fun.", envir = .GlobalEnv,
         function(val) readline("Test"))
  expect_pass_message(fuzz(".local_fun.", list(NULL)))
  })
})

test_that("check object returned", {
  testthat::skip_on_cran()

  funs <- c("list", "data.frame", "+")
  SW({
  expect_message(res <- fuzz(funs, list(NULL)),
                 "Functions will be searched in the global namespace")
  })
  expect_s3_class(res,
                  "cbtf")
  expect_named(res,
               c("runs", "funs", "package", "ignore_patterns", "ignore_warnings"))
  expect_length(res,
                length(funs))
  expect_s3_class(res[[1]],
                  "data.frame")
  expect_named(res$runs[[1]],
               c("res", "msg"))
  expect_equal(nrow(res[[1]]),
               length(funs))
  expect_equal(res$runs[[1]]$res,
               c("OK", "OK", "FAIL"))
  expect_equal(res$runs[[1]]$msg,
               c("", "", "invalid argument to unary operator"))
  expect_equal(res$package,
               NA)

  SW({
  res <- fuzz("ls", list(NA))
  })
  expect_fuzz_result(res,
                     "FAIL", "invalid object for 'as.environment'")

  SW({
  res <- fuzz("median", list(letters))
  })
  expect_fuzz_result(res,
                     "WARN", "argument is not numeric or logical: returning NA")

  ## newline in warnings
  assign(".local_fun.", envir = .GlobalEnv,
         function(arg) {
           warning("warning\n")
           arg
         })
  SW({
  res <- fuzz(".local_fun.", list(1))
  })
  expect_fuzz_result(res,
                     "WARN", "warning")

  ## multiple warnings
  assign(".local_fun.", envir = .GlobalEnv,
         function(arg) {
           warning("first warning")
           warning("second warning")
           arg
         })
  SW({
  res <- fuzz(".local_fun.", list(1))
  })
  expect_fuzz_result(res,
                     "WARN", "first warning | second warning")
  SW({
  res <- fuzz(".local_fun.", list(1), ignore_patterns = "second")
  })
  expect_fuzz_result(res,
                     "WARN", "first warning | second warning")
  SW({
  res <- fuzz(".local_fun.", list(1), ignore_patterns = "warning")
  })
  expect_fuzz_result(res,
                     "OK", "first warning | second warning")

  ## ignore warnings
  SW({
  res <- fuzz("ls", list(NA), ignore_warnings = TRUE)
  })
  expect_fuzz_result(res,
                     "FAIL", "invalid object for 'as.environment'")

  SW({
  res <- fuzz("median", list(letters), ignore_warnings = TRUE)
  })
  expect_fuzz_result(res,
                     "OK", "argument is not numeric or logical: returning NA")

  ## timeout
  assign(".local_fun.", envir = .GlobalEnv,
         function(arg) Sys.sleep(10))
  SW({
  res <- fuzz(".local_fun.", list(NA))
  })
  expect_fuzz_result(res,
                     "OK", "Timed out after 2 seconds")

  ## no timeouts
  assign(".local_fun.", envir = .GlobalEnv,
         function(arg) Sys.sleep(1))
  SW({
  res <- fuzz(rep(".local_fun.", 5), list(NA))
  })
  ## Windows CI produces: "" "Timed out after 2 seconds" "" "" ""
  if (.Platform$OS.type != "windows") {
  expect_fuzz_result(res,
                     rep("OK", 5), rep("", 5))
  }

  ## in case of both error and warning, we should report the error
  SW({
  assign(".local_fun.", envir = .GlobalEnv,
         function(arg) {
           warning("a warning")
           stop("an error", call. = FALSE)
         })
  res <- fuzz(".local_fun.", list(NA))
  })
  expect_fuzz_result(res,
                     "FAIL", "an error")

  ## check that we store the package attribute correctly
  funs <- structure(c("list", "data.frame"), package = "packagename")
  SW({
  res <- fuzz(funs, list(NULL))
  })
  expect_equal(res$package,
               "packagename")

  ## check that we don't generate spurious lists
  SW({
  assign(".local_fun.", envir = .GlobalEnv,
         function(arg) {
           is.list(arg) && stop("spurious list generated", call. = FALSE)
         })
  res <- fuzz(".local_fun.", list(NULL))
  })
  expect_fuzz_result(res,
                     "OK", "")

  ## test with the default inputs
  SW({
  res <- fuzz("list")
  })
  expect_length(res,
                length(test_inputs()))
  expect_equal(res$package,
               NA)
  SW({
  res <- fuzz("list", listify_what = TRUE)
  })
  expect_length(res,
                length(test_inputs()) * 2)

  ## NULL what
  SW({
  res <- fuzz("median", what = NULL, args = list(iris))
  })
  expect_length(res, 1)
  expect_fuzz_result(res,
                     "FAIL", "need numeric data")

  ## multiple arguments
  SW({
  res <- fuzz("median", what = list(NULL), args = list(1:5, TRUE))
  })
  expect_length(res, 2)

  SW({
  res <- fuzz("median", what = list(NULL, NA), args = list(x = 1:5, TRUE),
              listify_what = TRUE)
  })
  expect_length(res, 8)
  expect_equal(sapply(res$runs, attr, "what"),
               c("x = NULL, TRUE", "x = NA, TRUE",
                 "x = list(NULL), TRUE", "x = list(NA), TRUE",
                 "x = 1:5, NULL", "x = 1:5, NA",
                 "x = 1:5, list(NULL)", "x = 1:5, list(NA)"))

  SW({
  res <- fuzz("+", what = list(NULL), args = list(1:5, 2, NA))
  })
  expect_fuzz_result(res,
                     rep("SKIP", 3), rep("Accepts only up to 2 arguments", 3))
})

test_that("check classes returned", {
  testthat::skip_on_cran()

  SW({
  expect_what(fuzz("list", list(NULL)),
              "NULL")
  expect_what(fuzz("list", list(NA)),
              "NA")
  expect_what(fuzz("list", list(`data.frame()` = data.frame())),
              "data.frame()")
  expect_what(fuzz("list", list(list())),
              "list()")
  expect_what(fuzz("list", list(1:3)),
              "1:3")
  expect_what(fuzz("list", list(NA, letters = letters)),
              c("NA", "letters"))
  what <- letters
  expect_what(fuzz("list", list(input = what)),
              "input")
  })
})

test_that("self fuzz", {
  testthat::skip_on_cran()

  ## start nested daemons
  mirai::everywhere(mirai::daemons(1, dispatcher = FALSE))

  SW({
  expect_output(expect_pass_message(fuzz("fuzz", list(list())),
                                    "[fuzz] 'funs' should be of class character"),
                "OK 1")
  expect_output(expect_pass_message(fuzz("fuzz", list(NULL)),
                                    "[fuzz] 'funs' should be of class character"),
                "OK 1")

  ## fuzz test other arguments by currying the function
  curry_fuzz_for <- function(argname) {
    function(arg) do.call(fuzz, setNames(list("list", arg), c("funs", argname)))
  }
  test_self_fuzz <- function(argname, argtype) {
    assign(".local_fun.", envir = .GlobalEnv,
           curry_fuzz_for(argname))
    expect_pass_message(fuzz(".local_fun.",
                             ignore_patterns = "\\[fuzz\\]"),
                        sprintf("[fuzz] '%s' should be of class %s",
                                argname, paste(argtype, collapse = ", ")))
  }

  test_self_fuzz("package", "character")
  test_self_fuzz("listify_what", "logical")
  test_self_fuzz("ignore_patterns", "character")
  test_self_fuzz("ignore_warnings", "logical")
  test_self_fuzz("daemons", c("integer", "numeric"))
  test_self_fuzz("timeout", c("numeric", "integer"))

  ## as `what` expects a list argument, we can't use curry_fuzz_for()
  assign(".local_fun.", envir = .GlobalEnv,
         function(arg) fuzz("list", what = list(arg)))
  ## Windows CI produces: "Timed out after 2 seconds"
  if (.Platform$OS.type != "windows") {
  expect_pass_message(fuzz(".local_fun."))
  }
  })

  ## shut down nested daemons
  mirai::everywhere(mirai::daemons(0))
})

test_that("whitelist", {
  testthat::skip_on_cran()

  SW({
  res <- fuzz("numToInts")
  })
  expect_error(whitelist(NA, NA),
               "[whitelist] 'object' should be of class cbtf",
               fixed = TRUE)
  expect_error(whitelist(res, NA),
               "'patterns' should be of class character")
  expect_error(whitelist(res, ""),
               "'patterns' is an empty character")

  ignore_patterns <- c("cannot be coerced to type",
                       "imaginary parts discarded in coercion",
                       "NAs introduced by coercion")
  res.new <- whitelist(res, c("", ignore_patterns))
  SW({
  expect_pass_message(res.new)
  })
  expect_equal(res.new$ignore_patterns,
               ignore_patterns)
})

test_that("get_exported_functions", {
  testthat::skip_on_cran()

  expect_error(get_exported_functions(),
               "[get_exported_functions] 'package' should be of class character",
               fixed = TRUE)
  expect_error(get_exported_functions(NA),
               "'package' should be of class character")
  expect_error(get_exported_functions(character(0)),
               "'package' is an empty character")
  expect_error(get_exported_functions("like this"),
               "there is no package called 'like this'")
  expect_error(get_exported_functions("CBTF", NA),
               "'ignore_names' should be of class character")
  expect_error(get_exported_functions("CBTF", character(0)),
               "'ignore_names' is an empty character")
  expect_error(get_exported_functions("CBTF", "ignore", NA),
               "'ignore_deprecated' should be of class logical")
  expect_error(get_exported_functions("CBTF", "ignore", c(TRUE, FALSE)),
               "'ignore_deprecated' should be a single logical value")

  funs <- get_exported_functions("CBTF")
  expect_type(funs,
              "character")
  expect_equal(as.character(funs),
               c("fuzz",
                 "get_exported_functions",
                 "test_inputs",
                 "whitelist"))
  expect_equal(attr(funs, "package"),
               "CBTF")

  funs <- get_exported_functions("base")
  expect_false(".Device" %in% funs)
  expect_false("Sys.Date" %in% funs)

  SW({
  expect_pass_message(fuzz("get_exported_functions"),
                      "[get_exported_functions] 'package' should be of class character")
  assign(".local_fun.", envir = .GlobalEnv,
         function(arg) get_exported_functions(package = arg))
  expect_pass_message(fuzz(".local_fun.",
                           ignore_patterns = "\\[get_exported_functions\\]"),
                      "[get_exported_functions] 'package' should be of class character")
  })

  ## tested with mime 0.13
  skip_if_not_installed("mime")
  funs <- get_exported_functions("mime")
  expect_equal(as.character(funs),
               c("guess_type", "parse_multipart"))
})

test_that("regression test", {
  testthat::skip_on_cran()

  ## don't use SW as that would suppress warnings
  capture.output(suppressMessages(
  expect_pass_message(fuzz("median", list(NA),
                           timeout = .Machine$integer.max))
  ))
})

#' Default input tests
#'
#' This function provides a selection of potentially problematic inputs by
#' class. List inputs are very limited by design, as they can be automatically
#' generated by setting `listify_what = TRUE` in [fuzz].
#'
#' @param use Names of input classes to use. Valid names are "all" (default),
#'        "scalar", "numeric", "integer", "logical", "character", "factor",
#'        "data.frame", "matrix", "array", "date", "time", "raw", "na" and
#'        "list". A vector of valid classes can be retrieved programmatically
#'        by setting this argument to "help".
#' @param skip Names of input classes to skip. Overrides any classes selected
#'        via the `use` argument.
#'
#' @return
#' A named list of inputs corresponding to the input classes selected, or
#' a character vector of valid input classes if `use = "help"`.
#'
#' @examples
#' ## only the scalar and numeric tests
#' inputs1 <- test_inputs(use = c("scalar", "numeric"))
#'
#' ## everything but the data, raw and list tests
#' inputs2 <- test_inputs(skip = c("date", "raw", "list"))
#'
#' ## print the valid input classes
#' test_inputs("help")
#'
#' @seealso [fuzz]
#'
#' @export
test_inputs <- function(use = "all", skip = "") {
  validate_class(use, "character", remove_empty = TRUE)
  validate_class(skip, "character")
  inputs <- c("scalar_inputs",
              "numeric_inputs",
              "integer_inputs",
              "logical_inputs",
              "character_inputs",
              "factor_inputs",
              "data.frame_inputs",
              "matrix_inputs",
              "array_inputs",
              "date_inputs",
              "time_inputs",
              "raw_inputs",
              "na_inputs",
              "list_inputs")
  valid <- names(inputs) <- gsub("_inputs", "", inputs)
  "help" %in% use && return(c("all", valid))
  if ("all" %in% use)
    use <- valid
  use <- setdiff(use[use %in% valid], skip)
  length(use) == 0 &&
    fuzz_error("No valid tests selected, valid names are:",
               toString(sQuote(c("all", valid), q = FALSE)),
               from = "test_inputs")
  unlist(lapply(use, function(x) eval(call(inputs[x]))), recursive = FALSE)
}

#' Add names to a list of inputs
#'
#' This function can be used to generate automatically pretty names in a
#' list of custom input objects. This can improve the output, especially when
#' structures such as data frames, matrices, and more complex objects are
#' involved.
#'
#' @param ... Objects, possibly named.
#'
#' @return
#' A named list containing the evaluated arguments. For unnamed arguments,
#' names are generated by deparsing the unevaluated inputs.
#'
#' @examples
#' namify(data.frame(a = 1, b = 2))
#'
#' @seealso [fuzz]
#'
#' @noRd
namify <- function(...) {
  what <- substitute(list(...))

  ## use deparsed names for unnamed objects
  idx.empty <- if (is.null(names(what))) seq_along(what) else names(what) == ""
  names(what)[idx.empty] <- vapply(what, deparse, character(1), nlines = 1)[idx.empty]

  ## evaluate the arguments and remove the list() element added by substitute()
  lapply(what, eval)[-1]
}

scalar_inputs <- function() {
  namify(
      3L,
      0L,
      -4L,
      .Machine$integer.max,
      .Machine$double.xmin,
      .Machine$double.xmax,
      1.2,
      -2.5,
      6i,
      NaN,
      Inf,
      NULL
  )
}

numeric_inputs <- function() {
  namify(
      c(1.309605, 0.585381, -0.461072),
      c(-1, 0, NaN, 1e4),
      c(Inf, -0.5, 1234),
      c(0, 0),
      c(0, NA),
      numeric()
  )
}

integer_inputs <- function() {
  namify(
      -1:3,
      c(0L, NA),
      integer()
  )
}

logical_inputs <- function() {
  namify(
      TRUE,
      c(TRUE, NA, FALSE),
      logical()
  )
}

character_inputs <- function() {
  namify(
      "",
      "a test",
      c("", "B", ""),
      c("A", NA, "7", "+"),
      character()
  )
}

factor_inputs <- function() {
  namify(
      factor(""),
      factor(c("A", NA, "7", "+")),
      factor(c("a", NA, "c", "d"), levels = letters[1:4]),
      factor(c(NA, NA, NA, NA)),
      factor()
  )
}

date_inputs <- function() {
  namify(
      as.Date(NA),
      as.Date(NULL),
      as.Date(0),
      as.Date(c("2025-01-01", NA, "930-12-31")),
      Sys.Date()
  )
}

time_inputs <- function() {
  namify(
      as.POSIXct(NA),
      as.POSIXct(NULL),
      as.POSIXct("2065-01-01 00:00:00"),
      as.POSIXct(0, origin = "1970-01-01"),
      as.POSIXlt(Sys.time()),
      Sys.time()
  )
}

data.frame_inputs <- function() {
  namify(
      data.frame(a = NA),
      data.frame(a = letters),
      data.frame(a = 1:10, b = NA),
      data.frame(a = -1:1, b = Inf),
      data.frame(a = c(12, -2.1, 7), b = c(1.34, 2.28, 0.0), c = (-1:1 + 0.1)),
      datasets::iris[0, ],
      datasets::iris[, 0],
      datasets::iris[1, , drop = FALSE],
      datasets::iris[, 1, drop = FALSE],
      data.frame()
  )
}

matrix_inputs <- function() {
  namify(
      matrix(0, 0, 0),
      matrix(1, 1, 0),
      matrix(1, 0, 1),
      matrix(c(1.995874, 1.225707, -0.565287, -2.120309, 0.236326), 1, 5),
      matrix(c(-1.005770, 0.589365, 0.955131, -0.972982, 1.058721), 5, 1),
      matrix(c(1:7, NA, -1:-4), 3, 4),
      matrix(c(-1:6, Inf), 3, 3),
      matrix(letters, 13, 2),
      matrix()
  )
}

array_inputs <- function() {
  namify(
      array(0, c(0, 0, 0)),
      array(1, c(1, 1, 0)),
      array(1, c(0, 1, 0)),
      array(c(0.72350, -0.19754, 1.67188, -0.62163, -0.84149), c(1, 1, 5)),
      array(c(-0.16868, 0.86973, -0.46976, 0.32157, -0.43558), c(1, 5, 1)),
      array(c(-0.57441, -1.9585, -0.24757, -0.2467, -2.04617), c(5, 1, 1)),
      array(c(1:7, NA), c(2, 4, 1)),
      array(c(1:5, NA, NA), c(2, 4, 2)),
      array(c(1:3, Inf, 12.5), c(2, 4, 2)),
      array()
  )
}

raw_inputs <- function() {
  namify(
      charToRaw("0"),
      charToRaw(NA_character_),
      charToRaw("abc"),
      raw()
  )
}

na_inputs <- function() {
  namify(
      NA_real_,
      NA_integer_,
      NA_character_,
      NA
  )
}

list_inputs <- function() {
  namify(
      list(3),
      list(a = c(1, 2, 3), b = letters),
      list()
  )
}

test_that("test_inputs", {
  testthat::skip_on_cran()

  expect_error(test_inputs(NA),
               "'use' should be of class character")
  expect_error(test_inputs(""),
               "'use' is an empty character")
  expect_error(test_inputs("error"),
               "[test_inputs] No valid tests selected, valid names are: 'all'",
               fixed = TRUE)
  expect_error(test_inputs(skip = NA),
               "'skip' should be of class character")
  expect_error(test_inputs("scalar", skip = "scalar"),
               "No valid tests selected, valid names are: 'all', 'scalar'")

  expect_equal(test_inputs(use = "scalar", skip = "nonexisting"),
               test_inputs(use = "scalar"))
  expect_equal(test_inputs(use = c("scalar", "scalar")),
               test_inputs(use = "scalar"))
  expect_equal(test_inputs(use = c("scalar", "matrix"), skip = "matrix"),
               test_inputs(use = "scalar"))
  expect_equal(test_inputs(use = c("all", "scalar"), skip = "time"),
               test_inputs(skip = "time"))
  expect_equal(test_inputs(c("all", "error"), skip = "time"),
               test_inputs(skip = "time"))
  expect_equal(test_inputs("help"),
               c("all", "scalar", "numeric", "integer", "logical",
                 "character", "factor", "data.frame", "matrix", "array",
                 "date", "time", "raw", "na", "list"))
  expect_equal(test_inputs("help"),
               test_inputs(c("scalar", "help")))
})

## set up local daemons
mirai::daemons(2L)

## ensure that these functions are available when self-fuzzing
mirai::everywhere({},
                  fuzz = fuzz,
                  get_exported_functions = get_exported_functions)

## silence output and warnings
SW <- function(expr) capture.output(suppressMessages(suppressWarnings(expr)))

## check that the `res` and `msg` fields match the expectation
expect_fuzz_result <- function(res, exp_res, exp_msg) {
  expect_equal(c(sapply(res$runs, function(x) x$res)),
               exp_res)
  expect_equal(c(sapply(res$runs, function(x) x$msg)),
               exp_msg)
}

## check that the `what` attribute matches the expectation
expect_what <- function(res, exp) {
  expect_equal(sapply(res$runs, function(x) attr(x, "what")),
               exp)
}

## check that the function is skipped for the correct reason
expect_skip_reason <- function(res, reason) {
  expect_equal(paste(res$runs[[1]]$res, res$runs[[1]]$msg),
               paste("SKIP", reason))
}

## check that fuzzing found no errors
expect_pass_message <- function(res, msg = "") {
  expect_message(print(res),
                 "You didn't get caught by the fuzz!")
  expect_output(print(res),
                "FAIL 0 | WARN 0 | SKIP 0", fixed = TRUE)
  expect_equal(res$runs[[length(res$runs)]]$msg, msg)
}

## check that fuzzing found errors
expect_fail_message <- function(res) {
  expect_message(print(res),
                 "CAUGHT BY THE FUZZ!")
  expect_output(print(res),
                "SKIP 0 | OK 0", fixed = TRUE)
}

# summary

    Code
      summary(res)
    Message
      Fuzzed 1 function on 2 inputs:
    Output
            
             FAIL WARN SKIP OK
        list    0    0    0  2
      
      [ FAIL 0 | WARN 0 | SKIP 0 | OK 2 ]

---

    Code
      summary(res, tabulate = FALSE)
    Message
      Fuzzed 1 function on 2 inputs:
    Output
      
      [ FAIL 0 | WARN 0 | SKIP 0 | OK 2 ]

# print

    Code
      print(res)
    Message
      v    You didn't get caught by the fuzz!
    Output
      
       [ FAIL 0 | WARN 0 | SKIP 0 | OK 2 ] 

---

    Code
      print(res, group = "function", show = "all")
    Message
      v    You didn't get caught by the fuzz!
      
      -- Function `list`: 
    Output
         OK                                 | NA
         OK                                 | c(1, 2, 3)
      
       [ FAIL 0 | WARN 0 | SKIP 0 | OK 2 ] 

---

    Code
      print(res)
    Message
      x      CAUGHT BY THE FUZZ!    
      
      -- Test input [[2]]: 1:3, NA 
    Output
         median  FAIL  missing value where TRUE/FALSE needed
      
       [ FAIL 1 | WARN 0 | SKIP 2 | OK 3 ] 

---

    Code
      print(res, show = "all")
    Message
      x      CAUGHT BY THE FUZZ!    
      
      -- Test input [[1]]: NA, TRUE 
    Output
           list    OK  
         median    OK  
       Sys.date  SKIP  Object not found in the global namespace
    Message
      
      -- Test input [[2]]: 1:3, NA 
    Output
           list    OK  
         median  FAIL  missing value where TRUE/FALSE needed
       Sys.date  SKIP  Object not found in the global namespace
      
       [ FAIL 1 | WARN 0 | SKIP 2 | OK 3 ] 

---

    Code
      print(res, show = "skip")
    Message
      x      CAUGHT BY THE FUZZ!    
      
      -- Test input [[1]]: NA, TRUE 
    Output
       Sys.date  SKIP  Object not found in the global namespace
    Message
      
      -- Test input [[2]]: 1:3, NA 
    Output
       Sys.date  SKIP  Object not found in the global namespace
      
       [ FAIL 1 | WARN 0 | SKIP 2 | OK 3 ] 

---

    Code
      print(res, show = "none")
    Output
      [ FAIL 1 | WARN 0 | SKIP 2 | OK 3 ] 

---

    Code
      print(res, group = "function", show = "all")
    Message
      x      CAUGHT BY THE FUZZ!    
      
      -- Function `list`: 
    Output
         OK                                 | NA, TRUE
         OK                                 | 1:3, NA
    Message
      
      -- Function `median`: 
    Output
         OK                                        | NA, TRUE
       FAIL  missing value where TRUE/FALSE needed | 1:3, NA
    Message
      
      -- Function `Sys.date`: 
    Output
       SKIP  Object not found in the global namespace | NA, TRUE
       SKIP  Object not found in the global namespace | 1:3, NA
      
       [ FAIL 1 | WARN 0 | SKIP 2 | OK 3 ] 

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fuzz.R
\name{get_exported_functions}
\alias{get_exported_functions}
\title{Get the names of the exported functions of a package}
\usage{
get_exported_functions(package, ignore_names = "", ignore_deprecated = TRUE)
}
\arguments{
\item{package}{Name of the package where functions are searched.}

\item{ignore_names}{Names of functions to ignore: these are removed from
the names returned. This can be helpful, for example, to discard
function aliases.}

\item{ignore_deprecated}{Whether deprecated functions should be ignored
(\code{TRUE} by default).}
}
\value{
A character vector of the names of the fuzzable functions exported from
the given package, with the \code{"package"} attribute set. This can be used
directly as the \code{funs} argument of \link{fuzz} without the need to specify the
\code{package} argument.
}
\description{
This function extracts the exports from the namespace of the given package
via \link{getNamespaceExports} and discards non-fuzzable objects (non-functions
and functions with no arguments). The set of names returned can be further
restricted via the \code{ignore_names} and \code{ignore_deprecated} arguments.
}
\examples{
## get the fuzzable functions in the public interface of this package
funs <- get_exported_functions("CBTF")

}
\seealso{
\link{fuzz}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fuzz.R
\name{whitelist}
\alias{whitelist}
\title{Apply additional whitelist patterns to the results of a fuzz run}
\usage{
whitelist(object, patterns)
}
\arguments{
\item{object}{An object of class \code{cbtf}.}

\item{patterns}{One or more strings containing regular expressions to
match the errors to whitelist.}
}
\value{
An object of class \code{cbtf} with the additional whitelist patterns applied.
}
\description{
This allows for post-hoc whitelisting of results according to the patterns
specified.
}
\examples{

## this reports a false positive result
(res <- fuzz(funs = "matrix", what = test_inputs("scalar")))

## with whitelisting, we can remove that
whitelist(res, "must be of a vector type")

}
\seealso{
\link{fuzz}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.cbtf}
\alias{summary.cbtf}
\title{Results summary from a fuzz run}
\usage{
\method{summary}{cbtf}(object, tabulate = TRUE, ...)
}
\arguments{
\item{object}{An object of class \code{cbtf}.}

\item{tabulate}{Whether a tabulation of results should be printed out
(\code{TRUE} by default). The tabulation can always be retrieved from
the \code{"summary_table"} attribute of the returned object also when
\code{tabulate = FALSE}.}

\item{...}{Further arguments passed to or from other methods.
These are currently ignored.}
}
\value{
A data frame containing the following columns and attributes is returned
invisibly:
\item{fun}{The names of the function tested.}
\item{what}{The inputs tested.}
\item{res}{One of "OK", "FAIL", "WARN" or "SKIP" for each combination of
function and input tested (see the \emph{Value} section in \link{fuzz}).}
\item{msg}{The message received in case of error, warning or skip,
or an empty string if no failure occurred.}
\item{attr(*, "summary_table")}{The tabulation of results that was printed
out.}
}
\description{
Reports some summary statistics from the results of a run of \link{fuzz}.
}
\details{
The use of unicode icons in the output messages can be disabled by setting
\code{options(cli.unicode = FALSE)}.
}
\examples{
res <- fuzz(funs = c("list", "matrix", "mean"),
            what = test_inputs(c("numeric", "raw")))
summary(res)

}
\seealso{
\link{print.cbtf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{[[.cbtf}
\alias{[[.cbtf}
\title{Extract the results for a specific test input}
\usage{
\method{[[}{cbtf}(x, i)
}
\arguments{
\item{x}{An object of class \code{cbtf}.}

\item{i}{An index between 1 and the number of test inputs used.}
}
\value{
If the index is valid, a data frame containing the following columns and
attributes:
\item{res}{One of "OK", "FAIL", "WARN" or "SKIP" for each combination of
function and input tested (see the \emph{Value} section in \link{fuzz}).}
\item{msg}{The message received in case of error, warning or skip,
or an empty string if no failure occurred.}
\item{attr(*, "what")}{The character representation of the input tested.}
Otherwise, \code{FALSE}.
}
\description{
This extracts the raw results for a single test input by its position,
returning the corresponding data frame from the \verb{$runs} list.
}
\examples{
res <- fuzz(funs = c("list", "matrix", "mean"),
            what = test_inputs(c("numeric", "raw")))
res[[6]]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CBTF-package.R
\docType{package}
\name{CBTF-package}
\alias{CBTF}
\alias{CBTF-package}
\title{CBTF: Caught by the Fuzz! A minimalistic fuzz-test runner}
\description{
This package implements a very simple mechanism for fuzz-testing functions
in the public interface of an R package.
}
\details{
Fuzz testing helps identify functions that lack sufficient argument validation,
and uncovers sets of inputs that, while valid by function signature, may
cause issues within the function body.

The core functionality of the package is in \link{fuzz}, which calls each
provided function with a certain input and records the output produced.
If an error or a warning is generated, this is captured and reported to the
user, unless it matches a pattern of whitelisted messages. The objects
returned by \link{fuzz} can be inspected with \link{summary.cbtf} and \link{print.cbtf}.

Whitelisting can also be done after a fuzz run has been completed via the
\link{whitelist} function, so that only messages that need to be acted upon are
actually shown. Using \link{whitelist} has the advantage of not requiring the
completion of a fuzz run of all functions over all inputs again.

Note that \link{fuzz} relies on the \code{mirai} package for asynchronous operations
and parallelisation, and execution occurs on persistent background processes.
These can be started automatically by specifying the \code{daemons} option;
alternatively, they can be set up manually with the \code{mirai::daemons()}
function; refer to the original \code{mirai} documentation for a complete
description of its arguments and behaviour.

The helper function \link{get_exported_functions} identifies the functions
in the public interface of a given package, facilitating the generation of
the list of functions to be fuzzed.

The helper function \link{test_inputs} is invoked by \link{fuzz} if the user doesn't
specify the set of inputs to be tested. By default it generates a large set
of potentially problematic inputs, but these can be limited just to the
desired classes of inputs.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://mcol.github.io/caught-by-the-fuzz/}
  \item Report bugs at \url{https://github.com/mcol/caught-by-the-fuzz/issues}
}

}
\author{
\strong{Maintainer}: Marco Colombo \email{mar.colombo13@gmail.com} (\href{https://orcid.org/0000-0001-6672-0623}{ORCID})

Authors:
\itemize{
  \item Marco Colombo \email{mar.colombo13@gmail.com} (\href{https://orcid.org/0000-0001-6672-0623}{ORCID})
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{print.cbtf}
\alias{print.cbtf}
\title{Print the results from a fuzz run}
\usage{
\method{print}{cbtf}(x, show = c("fail", "warn"), group = "input", ...)
}
\arguments{
\item{x}{An object of class \code{cbtf}.}

\item{show}{A character vector representing the subset of results to be
printed, any of "fail", "warn", "skip", "ok", "all" and "none".}

\item{group}{Either \code{"input"} to show results grouped by test input
(default), or \code{"function"} to show them grouped by function name.}

\item{...}{Further arguments passed to or from other methods.
These are currently ignored.}
}
\value{
No return value, called for side effects.
}
\description{
This formats the results from a fuzz run with colours and prints them to
the terminal. By default, only tests labelled as \code{"FAIL"} or \code{"WARN"} are
displayed, but this can be controlled via the \code{show} argument.
}
\details{
The use of unicode icons in the output messages can be disabled by setting
\code{options(cli.unicode = FALSE)}.
}
\examples{
res <- fuzz(funs = c("list", "matrix", "mean"),
            what = test_inputs(c("numeric", "raw")))
print(res)
print(res, show = "all")
print(res, show = "none")
print(res, group = "function")

}
\seealso{
\link{summary.cbtf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{length.cbtf}
\alias{length.cbtf}
\title{Compute the number of tests performed}
\usage{
\method{length}{cbtf}(x)
}
\arguments{
\item{x}{An object of class \code{cbtf}.}
}
\value{
An integer corresponding to the number of tests performed in a run.
}
\description{
This returns the total number of tests as the product of the number of
inputs used and the number of functions tested.
}
\examples{
res <- fuzz(funs = c("list", "matrix", "mean"),
            what = test_inputs(c("numeric", "raw")))
length(res)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fuzz.R
\name{fuzz}
\alias{fuzz}
\title{Fuzz-test the specified functions}
\usage{
fuzz(
  funs,
  what = test_inputs(),
  args = NULL,
  package = NULL,
  listify_what = FALSE,
  ignore_patterns = "",
  ignore_warnings = FALSE,
  daemons = 2L,
  timeout = 2
)
}
\arguments{
\item{funs}{A character vector of function names to test. If the vector has
a \code{"package"} attribute and no \code{package} argument is given, functions
are loaded from the namespace specified in that attribute.}

\item{what}{A list of objects; each is used, in turn, to modify the list of
arguments in \code{args} before calling each of the functions in \code{funs}.
If no inputs are provided, a default set of inputs generated by
\link{test_inputs} will be used. If set to \code{NULL}, then \code{args} must be
specified, and all functions will be called with that exact list of
arguments with no fuzzing occurring.}

\item{args}{A list of arguments to pass to the functions being fuzzed. Each
element in the list is in turn replaced by each object in \code{what},
then each modified argument list is used to fuzz the functions in
\code{funs}. Argument names are preserved and used as named arguments in
the fuzzed functions. If \code{NULL} (default), only the first argument of
each function is fuzzed.}

\item{package}{Name of the package where functions are searched. If \code{NULL}
(default), the function first checks the \code{"package"} attribute of
\code{funs}, and if that is not set, names are searched in the global
namespace.}

\item{listify_what}{Whether each input in \code{what} should also be tested
in its listified version (\code{FALSE} by default). When set to \code{TRUE},
if \code{what} is \code{list(x = x)}, the function will operate as if it
were \code{list(x = x, "list(x)" = list(x))}, for any input object \code{x}.}

\item{ignore_patterns}{One or more strings containing regular expressions
to match the errors to ignore. The strings "unused argument" and
"is missing, with no default" are always ignored.}

\item{ignore_warnings}{Whether warnings should be ignored (\code{FALSE} by
default).}

\item{daemons}{Number of daemons to use (2 by default). As many \code{mirai}
daemons as specified will be started when entering the function and
closed at the end, unless active daemons are already available, in
which case the argument is ignored and the active daemons are used.}

\item{timeout}{Number of seconds (2 by default) after which the function
being fuzzed is interrupted with result status set to "OK".}
}
\value{
An object of class \code{cbtf} that stores the results obtained for each of the
functions tested. This contains the following fields:
\item{runs}{a list of data frames, each containing the results of fuzzing
all the functions in \code{funs} with one of the inputs in \code{what}. The
data frame contains the following columns and attributes:\cr
- \code{res}: The result of the fuzz test, see below for the possible
values.\cr
- \code{msg}: The error or warning message returned by the function, if
any.\cr
- \verb{attr(*, "what")}: The character representation of the input
tested.
}
\item{funs}{a vector of names of the functions tested.}
\item{args}{a named list of arguments, with names generated by deparsing
the \code{args} argument if not already provided.}
\item{package}{a character string specifying the package name where
function names were searched, or \code{NA} if none was provided.}
\item{ignore_patterns}{The value of the \code{ignore_patterns} argument.}
\item{ignore_warnings}{The value of the \code{ignore_warnings} argument.}

The \code{res} column in each of the data frames in the \verb{$runs} field can
contain the following values:
\itemize{
\item \strong{OK}: either no error or warning was produced (in which case, the \code{msg}
entry is left blank), or it was whitelisted (in which case, the message
received is stored in \code{msg}), or it was timed out (in which case, \code{msg}
records that a timeout was applied).
\item \strong{SKIP}: no test was run, either because the given name cannot be found, or
it doesn't correspond to a function, or the function accepts no arguments,
or more arguments were provided than the function accepts;
the exact reason is given in \code{msg}.
\item \strong{WARN}: a warning was thrown for which no whitelisting occurred and
\code{ignore_warnings = FALSE}; its message is stored in \code{msg}.
\item \strong{FAIL}: an error was thrown for which no whitelisting occurred; its message
is stored in \code{msg}.
}
}
\description{
The fuzzer calls each function in \code{funs} with the argument list provided in
\code{args} (each of its elements in turn modified by each object in \code{what}) and
records any errors or warnings that are thrown. If no error occurs within
the first \code{timeout} seconds, the execution of the function being fuzzed is
interrupted and the next one is started.
}
\details{
\subsection{Multiple arguments}{

A list of arguments to be passed to the functions being fuzzed can be
provided via the \code{args} argument. Each element in that list is modified in
turn by each object in \code{what} and the resulting list of arguments is then
passed to each function via \code{do.call()}. If more arguments are given than
the number of formal arguments accepted by a function, that function will
produce a "SKIP" result.

If arguments are named, they will be passed with their names to the fuzzed
functions. If a function doesn't have a formal argument of that name and
doesn't accept \code{...}, then the standard R behaviour is to return an "unused
argument" error. This is whitelisted by default in \code{fuzz()}, and the
corresponding result status is set to "OK".

It is possible to define arguments that should remain unchanged while
fuzzing by prefixing their names with \code{..}. These arguments will use the
values assigned in \code{args} without modification. For example, to ensure that
the argument \code{na.rm} is always set to \code{TRUE}, it should be specified as
\code{..na.rm = TRUE} in \code{args}. If all elements in \code{args} are fixed, \code{what} is
ignored and all functions in \code{funs} will be called with the provided \code{args}
list.
}

\subsection{Parallel execution}{

The implementation uses \code{mirai} as a backend to execute tasks asynchronously
in parallel worker processes. The function can start a pool of persistent
background processes (daemons) of size given by the \code{daemons} argument
(note that starting more daemons than available cores yields no benefit).
Alternatively, the function can also make use of already active daemons
started with the \link[mirai:daemons]{mirai::daemons} function: this allows to control in
greater detail the number of processes to use, which can also be remote.
}

\subsection{Whitelisting}{

In order to reduce the number of false positive results produced, this
function applies the following set of rules, to establish if an error or
warning condition should be ignored (whitelisting):
\itemize{
\item If the name of the function appears in the error or warning message, as
it is considered that the condition has been handled by the developer.
\item If the error or warning message contains the text "is missing, with no
default", which is produced when a missing argument is used without a
value being assigned to it.
\item If the error or warning message contains any of the patterns specified
in \code{ignore_patterns}. In case of multiple warnings being captured, the
result is whitelisted only if \emph{all} warnings match any of the patterns.
\item If a warning is thrown but \code{ignore_warnings = TRUE} is set.
}

In all whitelisted cases, the result is "OK", and the message that
was received is stored in the \verb{$msg} field (see the \emph{Value} section).

\emph{Note:} Whitelisting can also be applied post-hoc on the results of a fuzz
run using the \link{whitelist} function.
}
}
\examples{
## set up persistent background processes
mirai::daemons(2L)

## this should produce no errors
res <- fuzz(funs = c("list", "matrix", "mean"),
            what = test_inputs(c("numeric", "raw")))
summary(res)

## display all results even for successful tests
print(res, show = "all")

## this will catch an error (false positive)
fuzz(funs = "matrix",  what = test_inputs("scalar"))

## apply a whitelist pattern to remove the false positive
fuzz(funs = "matrix",  what = test_inputs("scalar"),
     ignore_patterns = "'data' must be of a vector type")

## close the background processes
mirai::daemons(0L)

}
\seealso{
\link{get_exported_functions}, \link{test_inputs}, \link{whitelist},
\link{summary.cbtf}, \link{print.cbtf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inputs.R
\name{test_inputs}
\alias{test_inputs}
\title{Default input tests}
\usage{
test_inputs(use = "all", skip = "")
}
\arguments{
\item{use}{Names of input classes to use. Valid names are "all" (default),
"scalar", "numeric", "integer", "logical", "character", "factor",
"data.frame", "matrix", "array", "date", "time", "raw", "na" and
"list". A vector of valid classes can be retrieved programmatically
by setting this argument to "help".}

\item{skip}{Names of input classes to skip. Overrides any classes selected
via the \code{use} argument.}
}
\value{
A named list of inputs corresponding to the input classes selected, or
a character vector of valid input classes if \code{use = "help"}.
}
\description{
This function provides a selection of potentially problematic inputs by
class. List inputs are very limited by design, as they can be automatically
generated by setting \code{listify_what = TRUE} in \link{fuzz}.
}
\examples{
## only the scalar and numeric tests
inputs1 <- test_inputs(use = c("scalar", "numeric"))

## everything but the data, raw and list tests
inputs2 <- test_inputs(skip = c("date", "raw", "list"))

## print the valid input classes
test_inputs("help")

}
\seealso{
\link{fuzz}
}
